"""
# Map legend components

Each map-legend component is either a 'mapnode' - something that represents and actual in-game
location (usually a room) or a 'maplink' - something connecting nodes together. The start of a link
usually shows as an Exit, but the length of the link has no in-game equivalent.

----

"""

try:
    from scipy import zeros
except ImportError as err:
    raise ImportError(
        f"{err}\nThe XYZgrid contrib requires the SciPy package. Install with `pip install scipy'."
    )

import uuid
from collections import defaultdict

from django.core import exceptions as django_exceptions

from evennia.prototypes import spawner
from evennia.utils.utils import class_from_module

from .utils import BIGVAL, MAPSCAN, REVERSE_DIRECTIONS, MapError, MapParserError

NodeTypeclass = None
ExitTypeclass = None


UUID_XYZ_NAMESPACE = uuid.uuid5(uuid.UUID(int=0), "xyzgrid")


# Nodes/Links


class MapNode:
    """
    This represents a 'room' node on the map. Note that the map system deals with two grids, the
    finer `xygrid`, which is the per-character grid on the map, and the `XYgrid` which contains only
    the even-integer coordinates and also represents in-game coordinates/rooms. MapNodes are always
    located on even X,Y coordinates on the map grid and in-game.

    MapNodes will also handle the syncing of themselves and all outgoing links to the grid.

    Attributes on the node class:

    - `symbol` (str) - The character to parse from the map into this node. By default this
      is '#' and must be a single character, with the exception of `\\
    - `display_symbol` (str or `None`) - This is what is used to visualize this node later. This
      symbol must still only have a visual size of 1, but you could e.g. use some fancy unicode
      character (be aware of encodings to different clients though) or, commonly, add color
      tags around it. For further customization, the `.get_display_symbol` method
      can return a dynamically determined display symbol. If set to `None`, the `symbol` is used.
    - `interrupt_path` (bool): If this is set, the shortest-path algorithm will include this
      node as normally, but the auto-stepper will stop when reaching it, even if not having reached
      its target yet. This is useful for marking 'points of interest' along a route, or places where
      you are not expected to be able to continue without some further in-game action not covered by
      the map (such as a guard or locked gate etc).
    - `prototype` (dict) - The default `prototype` dict to use for reproducing this map component
      on the game grid. This is used if not overridden specifically for this coordinate. If this
      is not given, nothing will be spawned for this coordinate (a 'virtual' node can be useful
      for various reasons, mostly map-transitions).

    """

    # symbol used to identify this link on the map
    symbol = "#"
    # if printing this node should show another symbol. If set
    # to the empty string, use `symbol`.
    display_symbol = None
    # this will interrupt a shortest-path step (useful for 'points' of interest, stop before
    # a door etc).
    interrupt_path = False
    # the prototype to use for mapping this to the grid.
    prototype = None

    # internal use. Set during generation, but is also used for identification of the node
    node_index = None
    # this should always be left True for Nodes and avoids inifinite loops during querying.
    multilink = True
    # default values to use if the exit doesn't have a 'spawn_aliases' iterable
    direction_spawn_defaults = {
        "n": ("north", "n"),
        "ne": ("northeast", "ne", "north-east"),
        "e": ("east", "e"),
        "se": ("southeast", "se", "south-east"),
        "s": ("south", "s"),
        "sw": ("southwest", "sw", "south-west"),
        "w": ("west", "w"),
        "nw": ("northwest", "nw", "north-west"),
        "d": ("down", "d", "do"),
        "u": ("up", "u"),
    }

    def __init__(self, x, y, Z, node_index=0, symbol=None, xymap=None):
        """
        Initialize the mapnode.

        Args:
            x (int): Coordinate on xygrid.
            y (int): Coordinate on xygrid.
            Z (int or str): Name/Z-pos of this map.
            node_index (int): This identifies this node with a running
                index number required for pathfinding. This is used
                internally and should not be set manually.
            symbol (str, optional): Set during parsing - allows to override
                the symbol based on what's set in the legend.
            xymap (XYMap, optional): The map object this sits on.

        """

        self.x = x
        self.y = y

        # map name, usually
        self.xymap = xymap

        # XYgrid coordinate
        self.X = x // 2
        self.Y = y // 2
        self.Z = Z

        self.node_index = node_index
        if symbol is not None:
            self.symbol = symbol

        # this indicates linkage in 8 cardinal directions on the string-map,
        # n,ne,e,se,s,sw,w,nw and link that to a node (always)
        self.links = {}
        # first MapLink in each direction - used by grid syncing
        self.first_links = {}
        # this maps
        self.weights = {}
        # lowest direction to a given neighbor
        self.shortest_route_to_node = {}
        # maps the directions (on the xygrid NOT on XYgrid!) taken if stepping
        # out from this node in a  given direction until you get to the end node.
        # This catches eventual longer link chains that would otherwise be lost
        # {startdirection: [direction, ...], ...}
        # where the directional path-lists also include the start-direction
        self.xy_steps_to_node = {}
        # direction-names of the closest neighbors to the node
        self.closest_neighbor_names = {}

    def __str__(self):
        return f"<MapNode '{self.symbol}' {self.node_index} XY=({self.X},{self.Y})"

    def __repr__(self):
        return str(self)

    def log(self, msg):
        """log messages using the xygrid parent"""
        self.xymap.log(msg)

    def generate_prototype_key(self):
        """
        Generate a deterministic prototype key to allow for users to apply prototypes without
        needing a separate new name for every one.

        """
        return str(uuid.uuid5(UUID_XYZ_NAMESPACE, str((self.X, self.Y, self.Z))))

    def build_links(self):
        """
        This is called by the map parser when this node is encountered. It tells the node
        to scan in all directions and follow any found links to other nodes. Since there
        could be multiple steps to reach another node, the system will iterate down each
        path and store it once and for all.

        Notes:
            This sets up all data needed for later use of this node in pathfinding and
            other operations. The method can't run immediately when the node is created
            since a complete parsed xygrid is required.

        """
        xygrid = self.xymap.xygrid

        # we must use the xygrid coordinates
        x, y = self.x, self.y

        # scan in all directions for links
        for direction, (dx, dy) in MAPSCAN.items():
            lx, ly = x + dx, y + dy

            if lx in xygrid and ly in xygrid[lx]:
                link = xygrid[lx][ly]

                # just because there is a link here, doesn't mean it has a
                # connection in this direction. If so, the `end_node` will be None.
                end_node, weight, steps = link.traverse(REVERSE_DIRECTIONS[direction])

                if end_node:
                    # the link could be followed to an end node!

                    self.first_links[direction] = link

                    # check the actual direction-alias to use, since this may be
                    # different than the xygrid cardinal directions. There must be
                    # no duplicates out of this node or there will be a
                    # multi-match error later!
                    first_step_name = steps[0].direction_aliases.get(direction, direction)
                    if first_step_name in self.closest_neighbor_names:
                        raise MapParserError(
                            f"has more than one outgoing direction '{first_step_name}'. "
                            "All directions out of a node must be unique.",
                            self,
                        )
                    self.closest_neighbor_names[first_step_name] = direction

                    node_index = end_node.node_index
                    self.weights[node_index] = weight
                    self.links[direction] = end_node
                    # this is useful for map building later - there could be multiple
                    # links tied together until getting to the node
                    self.xy_steps_to_node[direction] = steps

                    # used for building the shortest path. Note that we store the
                    # aliased link directions here, for quick display by the
                    # shortest-route solver
                    shortest_route = self.shortest_route_to_node.get(node_index, ("", [], BIGVAL))[
                        2
                    ]
                    if weight < shortest_route:
                        self.shortest_route_to_node[node_index] = (first_step_name, steps, weight)

    def linkweights(self, nnodes):
        """
        Retrieve all the weights for the direct links to all other nodes. This is
        used for the efficient generation of shortest-paths.

        Args:
            nnodes (int): The total number of nodes

        Returns:
            scipy.array: Array of weights of the direct links to other nodes.
                The weight will be 0 for nodes not directly connected to one another.

        Notes:
            A node can at most have 8 connections (the cardinal directions).

        """
        link_graph = zeros(nnodes)
        for node_index, weight in self.weights.items():
            link_graph[node_index] = weight
        return link_graph

    def get_display_symbol(self):
        """
        Hook to override for customizing how the display_symbol is determined.

        Returns:
            str: The display-symbol to use. This must visually be a single character
            but could have color markers, use a unicode font etc.

        Notes:
            By default, just setting .display_symbol is enough.

        """
        return self.symbol if self.display_symbol is None else self.display_symbol

    def get_spawn_xyz(self):
        """
        This should return the XYZ-coordinates for spawning this node. This normally
        the XYZ of the current map, but for traversal-nodes, it can also be the location
        on another map.

        Returns:
            tuple: The (X, Y, Z) coords to spawn this node at.
        """
        return self.X, self.Y, self.Z

    def get_exit_spawn_name(self, direction, return_aliases=True):
        """
        Retrieve the spawn name for the exit being created by this link.

        Args:
            direction (str): The cardinal direction (n,ne etc) the want the
                exit name/aliases for.
            return_aliases (bool, optional): Also return all aliases.

        Returns:
            str or tuple: The key of the spawned exit, or a tuple (key, alias, alias, ...)

        """
        key, *aliases = self.first_links[direction].spawn_aliases.get(
            direction, self.direction_spawn_defaults.get(direction, ("unknown",))
        )
        if return_aliases:
            return (key, *aliases)
        return key

    def spawn(self):
        """
        Build an actual in-game room from this node.

        This should be called as part of the node-sync step of the map sync. The reason is
        that the exits (next step) requires all nodes to exist before they can link up
        to their destinations.

        """
        global NodeTypeclass
        if not NodeTypeclass:
            from .xyzroom import XYZRoom as NodeTypeclass

        if not self.prototype:
            # no prototype means we can't spawn anything -
            # a 'virtual' node.
            return

        xyz = self.get_spawn_xyz()

        try:
            nodeobj = NodeTypeclass.objects.get_xyz(xyz=xyz)
        except django_exceptions.ObjectDoesNotExist:
            # create a new entity, using the specified typeclass (if there's one) and
            # with proper coordinates etc
            typeclass = self.prototype.get("typeclass")
            if typeclass is None:
                raise MapError(
                    f"The prototype {self.prototype} for this node has no 'typeclass' key.", self
                )
            self.log(f"  spawning room at xyz={xyz} ({typeclass})")
            Typeclass = class_from_module(typeclass)
            nodeobj, err = Typeclass.create(self.prototype.get("key", "An empty room"), xyz=xyz)
            if err:
                raise RuntimeError(err)
        else:
            self.log(f"  updating existing room (if changed) at xyz={xyz}")

        if not self.prototype.get("prototype_key"):
            # make sure there is a prototype_key in prototype
            self.prototype["prototype_key"] = self.generate_prototype_key()

        # apply prototype to node. This will not override the XYZ tags since
        # these are not in the prototype and exact=False
        spawner.batch_update_objects_with_prototype(self.prototype, objects=[nodeobj], exact=False)

    def spawn_links(self, directions=None):
        """
        Build actual in-game exits based on the links out of this room.

        Args:
            directions (list, optional): If given, this should be a list of supported
                directions (n, ne, etc). Only links in these directions will be spawned
                for this node.

        This should be called after all `sync_node_to_grid` operations have finished across
        the entire XYZgrid. This creates/syncs all exits to their locations and destinations.

        """
        if not self.prototype:
            # no exits to spawn out of a 'virtual' node.
            return

        xyz = (self.X, self.Y, self.Z)
        direction_limits = directions

        global ExitTypeclass
        if not ExitTypeclass:
            from .xyzroom import XYZExit as ExitTypeclass

        maplinks = {}
        for direction, link in self.first_links.items():
            key, *aliases = self.get_exit_spawn_name(direction)
            if not link.prototype.get("prototype_key"):
                # generate a deterministic prototype_key if it doesn't exist
                link.prototype["prototype_key"] = self.generate_prototype_key()
            maplinks[key.lower()] = (key, aliases, direction, link)

        # remove duplicates
        linkobjs = defaultdict(list)
        for exitobj in ExitTypeclass.objects.filter_xyz(xyz=xyz):
            linkobjs[exitobj.key].append(exitobj)
        for exitkey, exitobjs in linkobjs.items():
            for exitobj in exitobjs[1:]:
                self.log(f"  deleting duplicate {exitkey}")
                exitobj.delete()

        # we need to search for exits in all directions since some
        # may have been removed since last sync
        linkobjs = {exi.db_key.lower(): exi for exi in ExitTypeclass.objects.filter_xyz(xyz=xyz)}

        # figure out if the topology changed between grid and map (will always
        # build all exits first run)
        differing_keys = set(maplinks.keys()).symmetric_difference(set(linkobjs.keys()))
        for differing_key in differing_keys:
            if differing_key not in maplinks:
                # an exit without a maplink - delete the exit-object
                self.log(f"  deleting exit at xyz={xyz}, direction={differing_key}")

                linkobjs.pop(differing_key).delete()
            else:
                # missing in linkobjs - create a new exit
                key, aliases, direction, link = maplinks[differing_key]

                if direction_limits and direction not in direction_limits:
                    continue

                exitnode = self.links[direction]
                prot = maplinks[key.lower()][3].prototype
                typeclass = prot.get("typeclass")
                if typeclass is None:
                    raise MapError(
                        f"The prototype {prot} for this node has no 'typeclass' key.",
                        self,
                    )
                self.log(f"  spawning/updating exit xyz={xyz}, direction={key} ({typeclass})")

                Typeclass = class_from_module(typeclass)
                exi, err = Typeclass.create(
                    key,
                    xyz=xyz,
                    xyz_destination=exitnode.get_spawn_xyz(),
                    aliases=aliases,
                )
                if err:
                    raise RuntimeError(err)

                linkobjs[key.lower()] = exi

        # apply prototypes to catch any changes
        for key, linkobj in linkobjs.items():
            spawner.batch_update_objects_with_prototype(
                maplinks[key.lower()][3].prototype, objects=[linkobj], exact=False
            )

    def unspawn(self):
        """
        Remove all spawned objects related to this node and all links.

        """
        global NodeTypeclass
        if not NodeTypeclass:
            from .room import XYZRoom as NodeTypeclass

        xyz = (self.X, self.Y, self.Z)

        try:
            nodeobj = NodeTypeclass.objects.get_xyz(xyz=xyz)
        except django_exceptions.ObjectDoesNotExist:
            # no object exists
            pass
        else:
            nodeobj.delete()


class TransitionMapNode(MapNode):
    """
    This node acts as an end-node for a link that actually leads to a specific node on another
    map. It is not actually represented by a separate room in-game.

    This teleportation is not understood by the pathfinder, so why it will be possible to pathfind
    to this node, it really represents a map transition. Only a single link must ever be connected
    to this node.

    Properties:
    - `target_map_xyz` (tuple) - the (X, Y, Z) coordinate of a node on the other map to teleport
        to when moving to this node. This should not be another TransitionMapNode (see below for
        how to make a two-way link).

    Examples:
    ::

        map1   map2

        #-T    #-    - one-way transition from map1 -> map2.
        #-T    T-#   - two-way. Both TransitionMapNodes links to the coords of the
                       actual rooms (`#`) on the other map (NOT to the `T`s)!

    """

    symbol = "T"
    display_symbol = " "
    # X,Y,Z coordinates of target node
    taget_map_xyz = (None, None, None)

    def get_spawn_xyz(self):
        """
        Make sure to return the coord of the *target* - this will be used when building
        the exit to this node (since the prototype is None, this node itself will not be built).

        """
        if any(True for coord in self.target_map_xyz if coord in (None, "unset")):
            raise MapParserError(
                f"(Z={self.xymap.Z}) has not defined its "
                "`.target_map_xyz` property. It must point "
                "to another valid xymap (Z coordinate).",
                self,
            )

        return self.target_map_xyz

    def build_links(self):
        """Check so we don't have too many links"""
        super().build_links()
        if len(self.links) > 1:
            raise MapParserError("may have at most one link connecting to it.", self)


class MapLink:
    """
    This represents one or more links between an 'incoming direction'
    and an 'outgoing direction'. It's like a railway track between
    MapNodes. A Link can be placed on any location in the grid, but even when
    on an integer XY position they still don't represent an actual in-game place
    but just a link between such places (the Nodes).

    > Note that, if you want to create a child link-class that spawns onto the grid
    (usually an exit), you must set its `.prototype`. This parent class will not on
    its own spawn anything.

    Each link has a 'weight' >=1,  this indicates how 'slow'
    it is to traverse that link. This is used by the Dijkstra algorithm
    to find the 'fastest' route to a point. By default this weight is 1
    for every link, but a locked door, terrain etc could increase this
    and have the shortest-path algorithm prefer to use another route.

    Attributes on the link class:

    - `symbol` (str) - The character to parse from the map into this node. This must be a single
      character, with the exception of `\\`.
    - `display_symbol` (str or None)  - This is what is used to visualize this node later. This
      symbol must still only have a visual size of 1, but you could e.g. use some fancy unicode
      character (be aware of encodings to different clients though) or, commonly, add color
      tags around it. For further customization, the `.get_display_symbol` can be used.
    - `default_weight` (int) - Each link direction covered by this link can have its separate
      weight, this is used if none is specified in a particular direction. This value must be >= 1,
      and can be higher than 1 if a link should be less favored.
    - `directions` (dict) - this specifies which link edge to which other link-edge this link
      is connected; A link connecting the link's sw edge to its easted edge would be written
      as `{'sw': 'e'}` and read 'connects from southwest to east'. Note that if you want the
      link to go both ways, also the inverse (east to southwest) must also be added.
    - `weights (dict)` This maps a link's start direction to a weight. So for the
      `{'sw': 'e'}` link, a weight would be given as `{'sw': 2}`. If not given, a link will
      use the `default_weight`.
    - `average_long_link_weights` (bool): This applies to the *first* link out of a node only.
      When tracing links to another node, multiple links could be involved, each with a weight.
      So for a link chain with default weights, `#---#` would give a total weight of 3. With this
      setting, the weight will be 3 / 3 = 1. That is, for evenly weighted links, the length
      of the link doesn't matter.
    - `direction_aliases` (dict): When displaying a direction during pathfinding, one may want
      to display a different 'direction' than the cardinal on-map one. For example 'up' may be
      visualized on the map as a 'n' movement, but the found path over this link should show
      as 'u'. In that case, the alias would be `{'n': 'u'}`.
    - `multilink` (bool): If set, this link accepts links from all directions. It will usually
      use a custom get_direction to determine what these are based on surrounding topology. This
      setting is necessary to avoid infinite loops when such multilinks are next to each other.
    - `interrupt_path` (bool): If set, a shortest-path solution will include this link as normal,
      but will stop short of actually moving past this link.
    - `prototype` (dict) - The default `prototype` dict to use for reproducing this map component
      on the game grid. This is only relevant for the *first* link out of a Node (the continuation
      of the link is only used to determine its destination). This can be overridden on a
      per-direction basis.
    - `spawn_aliases` (dict): A mapping {direction: (key, alias, alias, ...) to use when spawning
      actual exits from this link. If not given, a sane set of defaults (n=(north, n) etc) will be
      used. This is required if you use any custom directions outside of the cardinal directions +
      up/down. The exit's key (useful for auto-walk) is usually retrieved by calling
      `node.get_exit_spawn_name(direction)`

    """

    # symbol for identifying this link on the map
    symbol = ""
    # if `None`, use .symbol
    display_symbol = None
    default_weight = 1
    # This setting only applies if this is the *first* link in a chain of multiple links. Usually,
    # when multiple links are used to tie together two nodes, the default is to average the weight
    # across all links. With this disabled, the weights will be added and a long link will be
    # considered 'longer' by the pathfinder.
    average_long_link_weights = True
    # this indicates linkage start:end in 8 cardinal directions on the string-map,
    # n,ne,e,se,s,sw,w,nw. A link is described as {startpos:endpoit}, like connecting
    # the named corners with a line. If the inverse direction is also possible, it
    # must also be specified. So a south-northward, two-way link would be described
    # as {"s": "n", "n": "s"}. The get_direction method can be customized to
    # return something else.
    directions = {}
    # for displaying the directions during pathfinding, you may want to show a different
    # direction than the cardinal one. For example, 'up' may be 'n' on the map, but
    # the direction when moving should be 'u'. This would be a alias {'n': 'u'}.
    direction_aliases = {}
    # this is required for pathfinding and contains cardinal directions (n, ne etc) only.
    # Each weight is defined as {startpos:weight}, where
    # the startpos is the direction of the cell (n,ne etc) where the link *starts*. The
    # weight is a value > 0, smaller than BIGVAL. The get_weight method can be
    # customized to modify to return something else.
    weights = {}
    # this shortcuts neighbors trying to figure out if they can connect to this link
    # - if this is set, they always can (similarly as to a node)
    multilink = False
    # this link does not block/reroute pathfinding, but makes the actual path always stop when
    # trying to cross it.
    interrupt_path = False
    # prototype for the first link out of a node.
    prototype = None
    # used for spawning and maps {direction: (key, alias, alias, ...) for use for the exits spawned
    # in this direction. Used unless the exit's prototype contain an explicit key - then that will
    # take precedence. If neither that nor this is not given, sane defaults ('n'=('north','n'), etc)
    # will be used.
    spawn_aliases = {}

    def __init__(self, x, y, Z, symbol=None, xymap=None):
        """
        Initialize the link.

        Args:
            x (int): The xygrid x coordinate
            y (int): The xygrid y coordinate.
            X (int or str): The name/Z-coord of this map we are on.
            symbol (str, optional): Set during parsing, allows to override
                the default symbol with the one set in the legend.
            xymap (XYMap, optional): The map object this sits on.

        """
        self.x = x
        self.y = y

        self.xymap = xymap

        self.X = x / 2
        self.Y = y / 2
        self.Z = Z

        if symbol is not None:
            self.symbol = symbol

    def __str__(self):
        return f"<LinkNode '{self.symbol}' XY=({self.X:g},{self.Y:g})>"

    def __repr__(self):
        return str(self)

    def generate_prototype_key(self, *args):
        """
        Generate a deterministic prototype key to allow for users to apply prototypes without
        needing a separate new name for every one.

        Args:
            *args (str): These are used to further seed the key.

        """
        return str(uuid.uuid5(UUID_XYZ_NAMESPACE, str((self.X, self.Y, self.Z, *args))))

    def traverse(self, start_direction, _weight=0, _linklen=1, _steps=None):
        """
        Recursively traverse the links out of this LinkNode.

        Args:
            start_direction (str): The direction (n, ne etc) from which
                this traversal originates for this link.
        Kwargs:
            _weight (int): Internal use.
            _linklen (int): Internal use.
            _steps (list): Internal use.

        Returns:
            tuple: The (node, weight, links) result of the traversal, where links
                is a list of directions (n, ne etc) that describes how to to get
                to the node on the grid. This includes the first direction.

        Raises:
            MapParserError: If a link lead to nowhere.

        """
        xygrid = self.xymap.xygrid

        end_direction = self.get_direction(start_direction)
        if not end_direction:
            if _steps is None:
                # is perfectly okay to not be linking back on the first step (to a node)
                return None, 0, None
            raise MapParserError(
                f"was connected to from the direction {start_direction}, but "
                "is not set up to link in that direction.",
                self,
            )

        # note that if `get_direction` returns an unknown direction, this will be equivalent
        # to pointing to an empty location, which makes sense
        dx, dy = MAPSCAN.get(end_direction, (BIGVAL, BIGVAL))
        end_x, end_y = self.x + dx, self.y + dy
        try:
            next_target = xygrid[end_x][end_y]
        except KeyError:
            # check if we have some special action up our sleeve
            next_target = self.at_empty_target(start_direction, end_direction)

        if not next_target:
            raise MapParserError(f"points to empty space in the direction {end_direction}!", self)

        _weight += self.get_weight(start_direction, _weight)
        if _steps is None:
            _steps = []
        _steps.append(self)

        if hasattr(next_target, "node_index"):
            # we reached a node, this is the end of the link.
            # we average the weight across all traversed link segments.
            return (
                next_target,
                _weight / max(1, _linklen) if self.average_long_link_weights else _weight,
                _steps,
            )
        else:
            # we hit another link. Progress recursively.
            return next_target.traverse(
                REVERSE_DIRECTIONS.get(end_direction, end_direction),
                _weight=_weight,
                _linklen=_linklen + 1,
                _steps=_steps,
            )

    def get_linked_neighbors(self, directions=None):
        """
        A helper to get all directions to which there appears to be a
        visual link/node. This does not trace the length of the link and check weights etc.

        Args:
            directions (list, optional): Only scan in these directions.

        Returns:
            dict: Mapping {direction: node_or_link} wherever such was found.

        """
        if not directions:
            directions = REVERSE_DIRECTIONS.keys()

        xygrid = self.xymap.xygrid
        links = {}
        for direction in directions:
            dx, dy = MAPSCAN[direction]
            end_x, end_y = self.x + dx, self.y + dy
            if end_x in xygrid and end_y in xygrid[end_x]:
                # there is is something there, we need to check if it is either
                # a map node or a link connecting in our direction
                node_or_link = xygrid[end_x][end_y]
                if node_or_link.multilink or node_or_link.get_direction(direction):
                    links[direction] = node_or_link
        return links

    def at_empty_target(self, start_direction, end_direction):
        """
        This is called by `.traverse` when it finds this link pointing to nowhere.

        Args:
            start_direction (str): The direction (n, ne etc) from which
                this traversal originates for this link.
            end_direction (str): The direction found from `get_direction` earlier.

        Returns:
            MapNode, MapLink or None: The next target to go to from here. `None` if this
            is an error that should be reported.

        Notes:
            This is usually a mapping error (returning `None`) but may have practical use, such as
            teleporting or transitioning to another map.

        """
        return None

    def get_direction(self, start_direction, **kwargs):
        """
        Hook to override for customizing how the directions are
        determined.

        Args:
            start_direction (str): The starting direction (n, ne etc).

        Returns:
            str: The 'out' direction side of the link - where the link
                leads to.

        Example:
            With the default legend, if the link is a straght vertical link
            (`|`) and `start_direction` is `s` (link is approached from
            from the south side), then this function will return `n'.

        """
        return self.directions.get(start_direction)

    def get_weight(self, start_direction, current_weight, **kwargs):
        """
        Hook to override for customizing how the weights are determined.

        Args:
            start_direction (str): The starting direction (n, ne etc).
            current_weight (int): This can have an existing value if
                we are progressing down a multi-step path.

        Returns:
            int: The weight to use for a link from `start_direction`.

        """
        return self.weights.get(start_direction, self.default_weight)

    def get_display_symbol(self):
        """
        Hook to override for customizing how the display_symbol is determined.
        This is called after all other hooks, at map visualization.

        Returns:
            str: The display-symbol to use. This must visually be a single character
            but could have color markers, use a unicode font etc.

        Notes:
            By default, just setting .display_symbol is enough.

        """
        return self.symbol if self.display_symbol is None else self.display_symbol


class SmartRerouterMapLink(MapLink):
    r"""
    A 'smart' link without visible direction, but which uses its topological surroundings
    to figure out how it connects. All such links are two-way. It can be used to create 'knees' and
    multi-crossings of links. Remember that this is still a link, so user will not 'stop' at it,
    even if placed on an XY position!

    If there are links on cardinally opposite sites, these are considered pass-throughs, and
    If determining the path of a set of input/output directions this is not possible, or there is an
    uneven number of links, an `MapParserError` is raised.

    Example with the RedirectLink:
    ::
          /
        -o    - this is ok, there can only be one path, e-ne

         |
        -o-   - equivalent to '+', one n-s and one w-e link crossing
         |

        \|/
        -o-   - all are passing straight through
        /|\

        -o-   - w-e pass straight through, other link is sw-s
        /|

        -o    - invalid; impossible to know which input goes to which output
        /|

    """
    multilink = True

    def get_direction(self, start_direction):
        """
        Dynamically determine the direction based on a source direction and grid topology.

        """
        # get all visually connected links
        if not self.directions:
            directions = {}
            unhandled_links = list(self.get_linked_neighbors().keys())

            # get all straight lines (n-s, sw-ne etc) we can trace through
            # the dynamic link and remove them from the unhandled_links list
            unhandled_links_copy = unhandled_links.copy()
            for direction in unhandled_links_copy:
                if REVERSE_DIRECTIONS[direction] in unhandled_links_copy:
                    directions[direction] = REVERSE_DIRECTIONS[
                        unhandled_links.pop(unhandled_links.index(direction))
                    ]

            # check if we have any non-cross-through paths left to handle
            n_unhandled = len(unhandled_links)
            if n_unhandled:
                # still remaining unhandled links. If there's not exactly
                # one 'incoming' and one 'outgoing' we can't figure out
                # where to go in a non-ambiguous way.
                if n_unhandled != 2:
                    links = ", ".join(unhandled_links)
                    raise MapParserError(
                        f"cannot determine how to connect in/out directions {links}.", self
                    )

                directions[unhandled_links[0]] = unhandled_links[1]
                directions[unhandled_links[1]] = unhandled_links[0]

            self.directions = directions

        return self.directions.get(start_direction)


class SmartTeleporterMapLink(MapLink):
    """
    The teleport link works by connecting to nowhere - and will then continue
    on another teleport link with the same symbol elsewhere on the map. The teleport
    symbol must connect to only one other link (not to a node).

    For this to work, there must be exactly one other teleport with the same `.symbol` on the map.
    The two teleports will always operate as two-way connections, but by making the 'out-link' on
    one side one-way, the effect will be that of a one-way teleport.

    Example:
    ::

          t #
         /  |   -  moving ne from the left node will bring the user to the rightmost node
       -#   t      as if the two teleporters were connected (two way).

       -#-t t># - one-way teleport from left to right.

       -#t       - invalid, may only connect to another link

       -#-t-#    - invalid, only one connected link is allowed.

    """

    symbol = "t"
    # usually invisible
    display_symbol = " "
    direction_name = "teleport"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.paired_teleporter = None

    def at_empty_target(self, start_direction, end_direction):
        """
        Called during traversal, when finding an unknown direction out of the link (same as
        targeting a link at an empty spot on the grid). This will also search for
        a unique, matching teleport to send to.

        Args:
            start_direction (str): The direction (n, ne etc) from which this traversal originates
                for this link.

        Returns:
            TeleporterMapLink: The paired teleporter.

        Raises:
            MapParserError: We raise this explicitly rather than returning `None` if we don't find
                another teleport. This avoids us getting the default (and in this case confusing)
                'pointing to an empty space' error we'd get if returning `None`.

        """
        xygrid = self.xymap.xygrid
        if not self.paired_teleporter:
            # scan for another teleporter
            symbol = self.symbol
            found_teleporters = []
            for iy, line in xygrid.items():
                for ix, node_or_link in xygrid[iy].items():
                    if node_or_link.symbol == symbol and node_or_link is not self:
                        found_teleporters.append(node_or_link)

            if not found_teleporters:
                raise MapParserError("found no matching teleporter to link to.", self)
            if len(found_teleporters) > 1:
                raise MapParserError(
                    "found too many matching teleporters (must be exactly one more): "
                    f"{found_teleporters}",
                    self,
                )

            other_teleporter = found_teleporters[0]
            # link the two so we don't need to scan again for the other one
            self.paired_teleporter = other_teleporter
            other_teleporter.paired_teleporter = self

        return self.paired_teleporter

    def get_direction(self, start_direction):
        """
        Figure out the connected link and paired teleport.

        """
        if not self.directions:
            neighbors = self.get_linked_neighbors()

            if len(neighbors) != 1:
                raise MapParserError("must have exactly one link connected to it.", self)
            direction, link = next(iter(neighbors.items()))
            if hasattr(link, "node_index"):
                raise MapParserError(
                    "can only connect to a Link. Found {link} in direction {direction}.", self
                )
            # the string 'teleport' will not be understood by the traverser, leading to
            # this being interpreted as an empty target and the `at_empty_target`
            # hook firing when trying to traverse this link.
            direction_name = self.direction_name
            if start_direction == direction_name:
                # called while traversing another teleport
                # - we must make sure we can always access/leave the teleport.
                self.directions = {direction_name: direction, direction: direction_name}
            else:
                # called while traversing a normal link
                self.directions = {start_direction: direction_name, direction_name: direction}

        return self.directions.get(start_direction)


class SmartMapLink(MapLink):
    """
    A 'smart' link withot visible direction, but which uses its topological surroundings
    to figure out how it connects. Unlike the `SmartRerouterMapLink`, this link type is
    also a 'direction' of its own and can thus connect directly to nodes. It can only describe
    one transition and will prefer connecting two nodes if there are other possibilities. If the
    linking is unclear or there are more than two nodes directly neighboring, a MapParserError will
    be raised.  If two nodes are not found, it will link to any combination of links- or nodes as
    long as it can un-ambiguously determine which direction they lead.

    Placing a smart-link directly between two nodes/links will always be a two-way connection,
    whereas if it connects a node with another link, it will be a one-way connection in the
    direction of the link.

    Example with the up-down directions:
    ::

        #
        u     - moving up in BOTH directions will bring you to the other node (two-way)
        #

        #
        d     - this better represents the 'real' up/down behavior.
        u
        #

        #
        |     - one-way up from the lower node to the upper
        u
        #

        #-#
        u     - okay since the up-link prioritizes the nodes
        #

        #u#
        u    - invalid since top-left node has two 'up' directions to go to
        #

        #     |
        u# or u-   - invalid.
        #     |

    """

    multilink = True

    def get_direction(self, start_direction):
        """
        Figure out the direction from a specific source direction based on grid topology.

        """
        # get all visually connected links
        if not self.directions:
            directions = {}
            neighbors = self.get_linked_neighbors()
            nodes = [
                direction
                for direction, neighbor in neighbors.items()
                if hasattr(neighbor, "node_index")
            ]

            if len(nodes) == 2:
                # prefer link to these two nodes
                for direction in nodes:
                    directions[direction] = REVERSE_DIRECTIONS[direction]
            elif len(neighbors) - len(nodes) == 1:
                for direction in neighbors:
                    directions[direction] = REVERSE_DIRECTIONS[direction]
            else:
                raise MapParserError(
                    "must have exactly two connections - either directly to "
                    "two nodes or connecting directly to one node and with exactly one other "
                    f"link direction. The neighbor(s) in directions {list(neighbors.keys())} do "
                    "not fulfill these criteria.",
                    self,
                )

            self.directions = directions
        return self.directions.get(start_direction)


class InvisibleSmartMapLink(SmartMapLink):
    """
    This is a smart maplink that does not show as such on the map - instead it will figure out
    how it should look had it been one of the 'normal' cardinal-direction links and display
    itself as that instead. This doesn't change its functionality, only the symbol shown
    on the map display. This only works for cardinal-direction links.

    It makes use of `display_symbol_aliases` mapping, which maps a sorted set of
    `((start, end), (end, start))` (two-way) or `((start, end),)` (one-way) directions
    to a symbol in the current map legend - this is the symbol alias to use. The matching
    MapLink or MapNode will be initialized at the current position only for the purpose of getting
    its display_symbol.

    Example:
        display_symbol_aliases = `{(('n', 's'), ('s', n')): '|', ...}`

    If no `display_symbol_aliases` are given, the regular display_symbol is used.

    """

    # this allows for normal movement directions even if the invisible-node
    # is marked with a different symbol.
    direction_aliases = {
        "n": "n",
        "ne": "ne",
        "e": "e",
        "se": "se",
        "s": "s",
        "sw": "sw",
        "w": "w",
        "nw": "nw",
    }

    # replace current link position with what the smart links "should" look like
    display_symbol_aliases = {
        (("n", "s"), ("s", "n")): "|",
        (("n", "s"),): "v",
        (("s", "n")): "^",
        (("e", "w"), ("w", "e")): "-",
        (("e", "w"),): ">",
        (("w", "e"),): "<",
        (("nw", "se"), ("sw", "ne")): "\\",
        (("ne", "sw"), ("sw", "ne")): "/",
    }

    def get_display_symbol(self):
        """
        The SmartMapLink already calculated the directions before this, so we
        just need to figure out what to replace this with in order to make this 'invisible'

        Depending on how we are connected, we figure out how the 'normal' link
        should look and use that instead.

        """
        if not hasattr(self, "_cached_display_symbol"):
            legend = self.xymap.legend
            default_symbol = self.symbol if self.display_symbol is None else self.display_symbol
            self._cached_display_symbol = default_symbol

            dirtuple = tuple((key, self.directions[key]) for key in sorted(self.directions.keys()))

            replacement_symbol = self.display_symbol_aliases.get(dirtuple, default_symbol)

            if replacement_symbol != self.symbol:
                node_or_link_class = legend.get(replacement_symbol)
                if node_or_link_class:
                    # initiate class in the current location and run get_display_symbol
                    # to get what it would show.
                    self._cached_display_symbol = node_or_link_class(
                        self.x, self.y, self.Z
                    ).get_display_symbol()
        return self._cached_display_symbol


# ----------------------------------
# Default nodes and link classes


class BasicMapNode(MapNode):
    """A map node/room"""

    symbol = "#"
    prototype = "xyz_room"


class InterruptMapNode(MapNode):
    """A point of interest node/room. Pathfinder will ignore but auto-stepper will
    stop here if passing through. Starting from here is fine."""

    symbol = "I"
    display_symbol = "#"
    interrupt_path = True
    prototype = "xyz_room"


class MapTransitionNode(TransitionMapNode):
    """Transition-target node to other map. This is not actually spawned in-game."""

    symbol = "T"
    display_symbol = " "
    prototype = None  # important to leave None!
    target_map_xyz = (None, None, None)  # must be set manually


class NSMapLink(MapLink):
    """Two-way, North-South link"""

    symbol = "|"
    display_symbol = "||"
    directions = {"n": "s", "s": "n"}
    prototype = "xyz_exit"


class EWMapLink(MapLink):
    """Two-way, East-West link"""

    symbol = "-"
    directions = {"e": "w", "w": "e"}
    prototype = "xyz_exit"


class NESWMapLink(MapLink):
    """Two-way, NorthWest-SouthWest link"""

    symbol = "/"
    directions = {"ne": "sw", "sw": "ne"}
    prototype = "xyz_exit"


class SENWMapLink(MapLink):
    """Two-way, SouthEast-NorthWest link"""

    symbol = "\\"
    directions = {"se": "nw", "nw": "se"}
    prototype = "xyz_exit"


class PlusMapLink(MapLink):
    """Two-way, crossing North-South and East-West links"""

    symbol = "+"
    directions = {"s": "n", "n": "s", "e": "w", "w": "e"}
    prototype = "xyz_exit"


class CrossMapLink(MapLink):
    """Two-way, crossing NorthEast-SouthWest and SouthEast-NorthWest links"""

    symbol = "x"
    directions = {"ne": "sw", "sw": "ne", "se": "nw", "nw": "se"}
    prototype = "xyz_exit"


class NSOneWayMapLink(MapLink):
    """One-way North-South link"""

    symbol = "v"
    directions = {"n": "s"}
    prototype = "xyz_exit"


class SNOneWayMapLink(MapLink):
    """One-way South-North link"""

    symbol = "^"
    directions = {"s": "n"}
    prototype = "xyz_exit"


class EWOneWayMapLink(MapLink):
    """One-way East-West link"""

    symbol = "<"
    directions = {"e": "w"}
    prototype = "xyz_exit"


class WEOneWayMapLink(MapLink):
    """One-way West-East link"""

    symbol = ">"
    directions = {"w": "e"}
    prototype = "xyz_exit"


class UpMapLink(SmartMapLink):
    """Up direction. Note that this stays on the same z-coord so it's a 'fake' up."""

    symbol = "u"

    # all movement over this link is 'up', regardless of where on the xygrid we move.
    direction_aliases = {
        "n": symbol,
        "ne": symbol,
        "e": symbol,
        "se": symbol,
        "s": symbol,
        "sw": symbol,
        "w": symbol,
        "nw": symbol,
    }
    spawn_aliases = {direction: ("up", "u") for direction in direction_aliases}
    prototype = "xyz_exit"


class DownMapLink(UpMapLink):
    """Down direction. Note that this stays on the same z-coord, so it's a 'fake' down."""

    symbol = "d"
    # all movement over this link is 'down', regardless of where on the xygrid we move.
    direction_aliases = {
        "n": symbol,
        "ne": symbol,
        "e": symbol,
        "se": symbol,
        "s": symbol,
        "sw": symbol,
        "w": symbol,
        "nw": symbol,
    }
    spawn_aliases = {direction: ("down", "d") for direction in direction_aliases}
    prototype = "xyz_exit"


class InterruptMapLink(InvisibleSmartMapLink):
    """A (still passable) link. Pathfinder will treat this as any link, but auto-stepper
    will always abort before crossing this link - so this must be crossed manually."""

    symbol = "i"
    interrupt_path = True
    prototype = "xyz_exit"


class BlockedMapLink(InvisibleSmartMapLink):
    """
    Causes the shortest-path algorithm to consider this a blocked path. The block will not show up
    in the map display (and exit can be traversed normally), pathfinder will just not include this
    link in any paths.

    """

    symbol = "b"
    weights = {
        "n": BIGVAL,
        "ne": BIGVAL,
        "e": BIGVAL,
        "se": BIGVAL,
        "s": BIGVAL,
        "sw": BIGVAL,
        "w": BIGVAL,
        "nw": BIGVAL,
    }
    prototype = "xyz_exit"


class RouterMapLink(SmartRerouterMapLink):
    """A link that connects other links to build 'knees', pass-throughs etc."""

    symbol = "o"


class TeleporterMapLink(SmartTeleporterMapLink):
    """
    Teleporter links. Must appear in pairs on the same xy map. To make it one-way, add additional
    one-way link out of the teleporter on one side.
    """

    symbol = "t"


# all map components; used as base if not overridden
LEGEND = {
    # nodes
    "#": BasicMapNode,
    "T": MapTransitionNode,
    "I": InterruptMapNode,
    # links
    "|": NSMapLink,
    "-": EWMapLink,
    "/": NESWMapLink,
    "\\": SENWMapLink,
    "x": CrossMapLink,
    "+": PlusMapLink,
    "v": NSOneWayMapLink,
    "^": SNOneWayMapLink,
    "<": EWOneWayMapLink,
    ">": WEOneWayMapLink,
    "o": RouterMapLink,
    "u": UpMapLink,
    "d": DownMapLink,
    "b": BlockedMapLink,
    "i": InterruptMapLink,
    "t": TeleporterMapLink,
}



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-01-28 23:11

from base64 import b64encode
from copy import deepcopy
from pickle import dumps, loads

from django.db import migrations

import evennia.utils.picklefield
from evennia.utils.utils import to_bytes


def forwards(apps, schema_editor):
    ServerConfig = apps.get_model("server", "ServerConfig")
    for conf in ServerConfig.objects.all():
        # picklefield requires base64 encoding
        value = loads(to_bytes(conf.db_value), encoding="bytes")  # py2->py3 byte encoding
        conf.db_value = b64encode(dumps(deepcopy(value), protocol=4)).decode()
        conf.save(update_fields=["db_value"])


class Migration(migrations.Migration):

    dependencies = [("server", "0001_initial")]

    operations = [
        migrations.RunPython(forwards, migrations.RunPython.noop),
        migrations.AlterField(
            model_name="serverconfig",
            name="db_value",
            field=evennia.utils.picklefield.PickledObjectField(
                help_text="The data returned when the config value is accessed. Must be written as a Python literal if editing through the admin interface. Attribute values which are not Python literals cannot be edited through the admin interface.",
                null=True,
                verbose_name="value",
            ),
        ),
    ]



"""
Evennia MUD/MUX/MU* creation system

This is the main top-level API for Evennia. You can explore the evennia library
by accessing evennia.<subpackage> directly. From inside the game you can read
docs of all object by viewing its `__doc__` string, such as through

    py evennia.ObjectDB.__doc__

For full functionality you should explore this module via a django-
aware shell. Go to your game directory and use the command

   evennia shell

to launch such a shell (using python or ipython depending on your install).
See www.evennia.com for full documentation.

"""
import evennia

# docstring header

DOCSTRING = """
Evennia MU* creation system.

Online manual and API docs are found at http://www.evennia.com.

Flat-API shortcut names:
{}
"""

# Delayed loading of properties

# Typeclasses

DefaultAccount = None
DefaultGuest = None
DefaultObject = None
DefaultCharacter = None
DefaultRoom = None
DefaultExit = None
DefaultChannel = None
DefaultScript = None

# Database models
ObjectDB = None
AccountDB = None
ScriptDB = None
ChannelDB = None
Msg = None
ServerConfig = None

# Properties
AttributeProperty = None
TagProperty = None
TagCategoryProperty = None

# commands
Command = None
CmdSet = None
default_cmds = None
syscmdkeys = None
InterruptCommand = None

# search functions
search_object = None
search_script = None
search_account = None
search_channel = None
search_message = None
search_help = None
search_tag = None

# create functions
create_object = None
create_script = None
create_account = None
create_channel = None
create_message = None
create_help_entry = None

# utilities
settings = None
lockfuncs = None
inputhandler = None
logger = None
gametime = None
ansi = None
spawn = None
managers = None
contrib = None
EvMenu = None
EvTable = None
EvForm = None
EvEditor = None
EvMore = None
ANSIString = None
signals = None
FuncParser = None
OnDemandTask = None

# Handlers
SESSION_HANDLER = None
PORTAL_SESSION_HANDLER = None
SERVER_SESSION_HANDLER = None
TASK_HANDLER = None
TICKER_HANDLER = None
MONITOR_HANDLER = None
ON_DEMAND_HANDLER = None

# Containers
GLOBAL_SCRIPTS = None
OPTION_CLASSES = None

PROCESS_ID = None

TWISTED_APPLICATION = None
EVENNIA_PORTAL_SERVICE = None
EVENNIA_SERVER_SERVICE = None


def _create_version():
    """
    Helper function for building the version string
    """
    import os
    from subprocess import STDOUT, CalledProcessError, check_output

    version = "Unknown"
    root = os.path.dirname(os.path.abspath(__file__))
    try:
        with open(os.path.join(root, "VERSION.txt"), "r") as f:
            version = f.read().strip()
    except IOError as err:
        print(err)
    try:
        rev = (
            check_output("git rev-parse --short HEAD", shell=True, cwd=root, stderr=STDOUT)
            .strip()
            .decode()
        )
        version = "%s (rev %s)" % (version, rev)
    except (IOError, CalledProcessError, OSError):
        # ignore if we cannot get to git
        pass
    return version


__version__ = _create_version()
del _create_version

_LOADED = False

PORTAL_MODE = False


def _init(portal_mode=False):
    """
    This function is called automatically by the launcher only after
    Evennia has fully initialized all its models. It sets up the API
    in a safe environment where all models are available already.
    """
    global _LOADED
    if _LOADED:
        return
    _LOADED = True
    global DefaultAccount, DefaultObject, DefaultGuest, DefaultCharacter
    global DefaultRoom, DefaultExit, DefaultChannel, DefaultScript
    global ObjectDB, AccountDB, ScriptDB, ChannelDB, Msg
    global Command, CmdSet, default_cmds, syscmdkeys, InterruptCommand
    global search_object, search_script, search_account, search_channel
    global search_help, search_tag, search_message
    global create_object, create_script, create_account, create_channel
    global create_message, create_help_entry, OnDemandTask
    global signals
    global settings, lockfuncs, logger, utils, gametime, ansi, spawn, managers
    global contrib, TICKER_HANDLER, MONITOR_HANDLER, SESSION_HANDLER, PROCESS_ID
    global TASK_HANDLER, PORTAL_SESSION_HANDLER, SERVER_SESSION_HANDLER, ON_DEMAND_HANDLER
    global GLOBAL_SCRIPTS, OPTION_CLASSES, EVENNIA_PORTAL_SERVICE, EVENNIA_SERVER_SERVICE, TWISTED_APPLICATION
    global EvMenu, EvTable, EvForm, EvMore, EvEditor
    global ANSIString, FuncParser
    global AttributeProperty, TagProperty, TagCategoryProperty, ServerConfig
    global PORTAL_MODE
    PORTAL_MODE = portal_mode

    # Parent typeclasses
    # utilities
    import os

    from django.conf import settings

    from . import contrib
    from .accounts.accounts import DefaultAccount, DefaultGuest
    from .accounts.models import AccountDB
    from .commands.cmdset import CmdSet
    from .commands.command import Command, InterruptCommand
    from .comms.comms import DefaultChannel
    from .comms.models import ChannelDB, Msg
    from .locks import lockfuncs
    from .objects.models import ObjectDB
    from .objects.objects import (
        DefaultCharacter,
        DefaultExit,
        DefaultObject,
        DefaultRoom,
    )
    from .prototypes.spawner import spawn
    from .scripts.models import ScriptDB
    from .scripts.monitorhandler import MONITOR_HANDLER
    from .scripts.ondemandhandler import ON_DEMAND_HANDLER, OnDemandTask
    from .scripts.scripts import DefaultScript
    from .scripts.taskhandler import TASK_HANDLER
    from .scripts.tickerhandler import TICKER_HANDLER
    from .server import signals
    from .server.models import ServerConfig
    from .typeclasses.attributes import AttributeProperty
    from .typeclasses.tags import TagCategoryProperty, TagProperty
    from .utils import ansi, class_from_module, gametime, logger
    from .utils.ansi import ANSIString

    if not PORTAL_MODE:
        # containers
        from .utils.containers import GLOBAL_SCRIPTS, OPTION_CLASSES

    # create functions
    from .utils.create import (
        create_account,
        create_channel,
        create_help_entry,
        create_message,
        create_object,
        create_script,
    )
    from .utils.eveditor import EvEditor
    from .utils.evform import EvForm
    from .utils.evmenu import EvMenu
    from .utils.evmore import EvMore
    from .utils.evtable import EvTable
    from .utils.funcparser import FuncParser

    # search functions
    from .utils.search import (
        search_account,
        search_channel,
        search_help,
        search_message,
        search_object,
        search_script,
        search_tag,
    )
    from .utils.utils import class_from_module

    PROCESS_ID = os.getpid()

    from twisted.application.service import Application

    TWISTED_APPLICATION = Application("Evennia")

    _evennia_service_class = None

    if portal_mode:
        # Set up the PortalSessionHandler
        from evennia.server.portal import portalsessionhandler

        portal_sess_handler_class = class_from_module(settings.PORTAL_SESSION_HANDLER_CLASS)
        portalsessionhandler.PORTAL_SESSIONS = portal_sess_handler_class()
        SESSION_HANDLER = portalsessionhandler.PORTAL_SESSIONS
        evennia.PORTAL_SESSION_HANDLER = evennia.SESSION_HANDLER
        _evennia_service_class = class_from_module(settings.EVENNIA_PORTAL_SERVICE_CLASS)
        EVENNIA_PORTAL_SERVICE = _evennia_service_class()
        EVENNIA_PORTAL_SERVICE.setServiceParent(TWISTED_APPLICATION)

        from django.db import connection

        # we don't need a connection to the database so close it right away
        try:
            connection.close()
        except Exception:
            pass

    else:
        # Create the ServerSesssionHandler
        from evennia.server import sessionhandler

        sess_handler_class = class_from_module(settings.SERVER_SESSION_HANDLER_CLASS)
        sessionhandler.SESSIONS = sess_handler_class()
        sessionhandler.SESSION_HANDLER = sessionhandler.SESSIONS
        SESSION_HANDLER = sessionhandler.SESSIONS
        SERVER_SESSION_HANDLER = SESSION_HANDLER
        _evennia_service_class = class_from_module(settings.EVENNIA_SERVER_SERVICE_CLASS)
        EVENNIA_SERVER_SERVICE = _evennia_service_class()
        EVENNIA_SERVER_SERVICE.setServiceParent(TWISTED_APPLICATION)

    # API containers

    class _EvContainer(object):
        """
        Parent for other containers

        """

        def _help(self):
            "Returns list of contents"
            names = [name for name in self.__class__.__dict__ if not name.startswith("_")]
            names += [name for name in self.__dict__ if not name.startswith("_")]
            print(self.__doc__ + "-" * 60 + "\n" + ", ".join(names))

        help = property(_help)

    class DBmanagers(_EvContainer):
        """
        Links to instantiated Django database managers. These are used
        to perform more advanced custom database queries than the standard
        search functions allow.

        helpentries - HelpEntry.objects
        accounts - AccountDB.objects
        scripts - ScriptDB.objects
        msgs    - Msg.objects
        channels - Channel.objects
        objects - ObjectDB.objects
        serverconfigs - ServerConfig.objects
        tags - Tags.objects
        attributes - Attributes.objects

        """

        from .accounts.models import AccountDB
        from .comms.models import ChannelDB, Msg
        from .help.models import HelpEntry
        from .objects.models import ObjectDB
        from .scripts.models import ScriptDB
        from .server.models import ServerConfig
        from .typeclasses.attributes import Attribute
        from .typeclasses.tags import Tag

        # create container's properties
        helpentries = HelpEntry.objects
        accounts = AccountDB.objects
        scripts = ScriptDB.objects
        msgs = Msg.objects
        channels = ChannelDB.objects
        objects = ObjectDB.objects
        serverconfigs = ServerConfig.objects
        attributes = Attribute.objects
        tags = Tag.objects
        # remove these so they are not visible as properties
        del HelpEntry, AccountDB, ScriptDB, Msg, ChannelDB
        # del ExternalChannelConnection
        del ObjectDB, ServerConfig, Tag, Attribute

    managers = DBmanagers()
    del DBmanagers

    class DefaultCmds(_EvContainer):
        """
        This container holds direct shortcuts to all default commands in Evennia.

        To access in code, do 'from evennia import default_cmds' then
        access the properties on the imported default_cmds object.

        """

        from .commands.default.cmdset_account import AccountCmdSet
        from .commands.default.cmdset_character import CharacterCmdSet
        from .commands.default.cmdset_session import SessionCmdSet
        from .commands.default.cmdset_unloggedin import UnloggedinCmdSet
        from .commands.default.muxcommand import MuxAccountCommand, MuxCommand

        def __init__(self):
            "populate the object with commands"

            def add_cmds(module):
                "helper method for populating this object with cmds"
                from evennia.utils import utils

                cmdlist = utils.variable_from_module(module, module.__all__)
                self.__dict__.update(dict([(c.__name__, c) for c in cmdlist]))

            from .commands.default import (
                account,
                admin,
                batchprocess,
                building,
                comms,
                general,
                help,
                system,
                unloggedin,
            )

            add_cmds(admin)
            add_cmds(building)
            add_cmds(batchprocess)
            add_cmds(building)
            add_cmds(comms)
            add_cmds(general)
            add_cmds(account)
            add_cmds(help)
            add_cmds(system)
            add_cmds(unloggedin)

    default_cmds = DefaultCmds()
    del DefaultCmds

    class SystemCmds(_EvContainer):
        """
        Creating commands with keys set to these constants will make
        them system commands called as a replacement by the parser when
        special situations occur. If not defined, the hard-coded
        responses in the server are used.

        CMD_NOINPUT - no input was given on command line
        CMD_NOMATCH - no valid command key was found
        CMD_MULTIMATCH - multiple command matches were found
        CMD_LOGINSTART - this command will be called as the very
                         first command when an account connects to
                         the server.

        To access in code, do 'from evennia import syscmdkeys' then
        access the properties on the imported syscmdkeys object.

        """

        from .commands import cmdhandler

        CMD_NOINPUT = cmdhandler.CMD_NOINPUT
        CMD_NOMATCH = cmdhandler.CMD_NOMATCH
        CMD_MULTIMATCH = cmdhandler.CMD_MULTIMATCH
        CMD_LOGINSTART = cmdhandler.CMD_LOGINSTART
        del cmdhandler

    syscmdkeys = SystemCmds()
    del SystemCmds
    del _EvContainer


def set_trace(term_size=(140, 80), debugger="auto"):
    """
    Helper function for running a debugger inside the Evennia event loop.

    Args:
        term_size (tuple, optional): Only used for Pudb and defines the size of the terminal
            (width, height) in number of characters.
        debugger (str, optional): One of 'auto', 'pdb' or 'pudb'. Pdb is the standard debugger. Pudb
            is an external package with a different, more 'graphical', ncurses-based UI. With
            'auto', will use pudb if possible, otherwise fall back to pdb. Pudb is available through
            `pip install pudb`.

    Notes:
        To use:

        1) add this to a line to act as a breakpoint for entering the debugger:

            from evennia import set_trace; set_trace()

        2) restart evennia in interactive mode

            evennia istart

        3) debugger will appear in the interactive terminal when breakpoint is reached. Exit
           with 'q', remove the break line and restart server when finished.

    """
    import sys

    dbg = None

    if debugger in ("auto", "pudb"):
        try:
            from pudb import debugger

            dbg = debugger.Debugger(stdout=sys.__stdout__, term_size=term_size)
        except ImportError:
            if debugger == "pudb":
                raise
            pass

    if not dbg:
        import pdb

        dbg = pdb.Pdb(stdout=sys.__stdout__)

    try:
        # Start debugger, forcing it up one stack frame (otherwise `set_trace`
        # will start debugger this point, not the actual code location)
        dbg.set_trace(sys._getframe().f_back)
    except Exception:
        # Stopped at breakpoint. Press 'n' to continue into the code.
        dbg.set_trace()


# initialize the doc string
global __doc__
__doc__ = DOCSTRING.format(
    "\n- "
    + "\n- ".join(
        f"evennia.{key}"
        for key in sorted(globals())
        if not key.startswith("_") and key not in ("DOCSTRING",)
    )
)



"""

Test the funcparser module.

"""

import time
import unittest
from ast import literal_eval
from unittest.mock import MagicMock, patch

from django.test import TestCase, override_settings
from parameterized import parameterized
from simpleeval import simple_eval

from evennia.utils import funcparser, test_resources


def _test_callable(*args, **kwargs):
    kwargs.pop("funcparser", None)
    kwargs.pop("raise_errors", None)
    argstr = ", ".join(args)
    kwargstr = ""
    if kwargs:
        kwargstr = (", " if args else "") + ", ".join(f"{key}={val}" for key, val in kwargs.items())
    return f"_test({argstr}{kwargstr})"


def _repl_callable(*args, **kwargs):
    if args:
        return f"r{args[0]}r"
    return "rr"


def _double_callable(*args, **kwargs):
    if args:
        try:
            return int(args[0]) * 2
        except ValueError:
            pass
    return "N/A"


def _eval_callable(*args, **kwargs):
    if args:
        return simple_eval(args[0])

    return ""


def _clr_callable(*args, **kwargs):
    clr, string, *rest = args
    return f"|{clr}{string}|n"


def _typ_callable(*args, **kwargs):
    try:
        if isinstance(args[0], str):
            return type(literal_eval(args[0]))
        else:
            return type(args[0])
    except (SyntaxError, ValueError):
        return type("")


def _add_callable(*args, **kwargs):
    if len(args) > 1:
        return literal_eval(args[0]) + literal_eval(args[1])
    return ""


def _lit_callable(*args, **kwargs):
    return literal_eval(args[0])


def _lsum_callable(*args, **kwargs):
    if isinstance(args[0], (list, tuple)):
        return sum(val for val in args[0])
    return ""


def _raises_callable(*args, **kwargs):
    raise RuntimeError("Test exception raised by test callable")


def _pass_callable(*args, **kwargs):
    kwargs.pop("funcparser", None)
    kwargs.pop("raise_errors", None)
    return str(args) + str(kwargs)


_test_callables = {
    "foo": _test_callable,
    "bar": _test_callable,
    "with spaces": _test_callable,
    "repl": _repl_callable,
    "double": _double_callable,
    "eval": _eval_callable,
    "clr": _clr_callable,
    "typ": _typ_callable,
    "add": _add_callable,
    "lit": _lit_callable,
    "sum": _lsum_callable,
    "raise": _raises_callable,
    "pass": _pass_callable,
}


class TestFuncParser(TestCase):
    """
    Test the FuncParser class

    """

    def setUp(self):
        self.parser = funcparser.FuncParser(_test_callables)

    def test_constructor_wrong_args(self):
        # Given list argument doesn't contain modules or paths.
        with self.assertRaises(AttributeError):
            parser = funcparser.FuncParser(["foo", _test_callable])

    def test_constructor_ignore_non_callables(self):
        # Ignores callables that aren't actual functions.
        parser = funcparser.FuncParser({"foo": 1, "bar": "baz"})

    @patch("evennia.utils.funcparser.variable_from_module")
    def test_constructor_raises(self, patched_variable_from_module):
        # Patched variable from module returns FUNCPARSER_CALLABLES that isn't dict.
        patched_variable_from_module.return_value = ["foo"]
        with self.assertRaises(funcparser.ParsingError):
            parser = funcparser.FuncParser("foo.module")

    @parameterized.expand(
        [
            ("Test normal string", "Test normal string"),
            ("Test noargs1 $foo()", "Test noargs1 _test()"),
            ("Test noargs2 $bar() etc.", "Test noargs2 _test() etc."),
            ("Test noargs3 $with spaces() etc.", "Test noargs3 _test() etc."),
            ("Test noargs4 $foo(), $bar() and $foo", "Test noargs4 _test(), _test() and $foo"),
            ("$foo() Test noargs5", "_test() Test noargs5"),
            ("Test args1 $foo(a,b,c)", "Test args1 _test(a, b, c)"),
            ("Test args2 $bar(foo, bar,    too)", "Test args2 _test(foo, bar, too)"),
            (r'Test args3 $bar(foo, bar, "   too")', "Test args3 _test(foo, bar,    too)"),
            ("Test args4 $foo('')", "Test args4 _test('')"),  # ' treated as literal
            ('Test args4 $foo("")', "Test args4 _test()"),
            ("Test args5 $foo(\(\))", "Test args5 _test(())"),
            ("Test args6 $foo(\()", "Test args6 _test(()"),
            ("Test args7 $foo(())", "Test args7 _test(())"),
            ("Test args8 $foo())", "Test args8 _test())"),
            ("Test args9 $foo(=)", "Test args9 _test(=)"),
            ("Test args10 $foo(\,)", "Test args10 _test(,)"),
            (r'Test args10 $foo(",")', "Test args10 _test(,)"),
            ("Test args11 $foo(()", "Test args11 $foo(()"),  # invalid syntax
            (
                r'Test kwarg1 $bar(foo=1, bar="foo", too=ere)',
                "Test kwarg1 _test(foo=1, bar=foo, too=ere)",
            ),
            ("Test kwarg2 $bar(foo,bar,too=ere)", "Test kwarg2 _test(foo, bar, too=ere)"),
            ("test kwarg3 $foo(foo = bar, bar = ere )", "test kwarg3 _test(foo=bar, bar=ere)"),
            (
                r"test kwarg4 $foo(foo =' bar ',\" bar \"= ere )",
                "test kwarg4 _test(foo=' bar ', \" bar \"=ere)",
            ),
            (
                "Test nest1 $foo($bar(foo,bar,too=ere))",
                "Test nest1 _test(_test(foo, bar, too=ere))",
            ),
            (
                "Test nest2 $foo(bar,$repl(a),$repl()=$repl(),a=b) etc",
                "Test nest2 _test(bar, rar, rr=rr, a=b) etc",
            ),
            ("Test nest3 $foo(bar,$repl($repl($repl(c))))", "Test nest3 _test(bar, rrrcrrr)"),
            (
                "Test nest4 $foo($bar(a,b),$bar(a,$repl()),$bar())",
                "Test nest4 _test(_test(a, b), _test(a, rr), _test())",
            ),
            ("Test escape1 \\$repl(foo)", "Test escape1 $repl(foo)"),
            (
                'Test escape2 "This is $foo() and $bar($bar())", $repl()',
                'Test escape2 "This is _test() and _test(_test())", rr',
            ),
            (
                "Test escape3 'This is $foo() and $bar($bar())', $repl()",
                "Test escape3 'This is _test() and _test(_test())', rr",
            ),
            (
                "Test escape4 $$foo() and $$bar(a,b), $repl()",
                "Test escape4 $foo() and $bar(a,b), rr",
            ),
            ("Test with color |r$foo(a,b)|n is ok", "Test with color |r_test(a, b)|n is ok"),
            ("Test malformed1 This is $foo( and $bar(", "Test malformed1 This is $foo( and $bar("),
            (
                "Test malformed2 This is $foo( and  $bar()",
                "Test malformed2 This is $foo( and  _test()",
            ),
            ("Test malformed3 $", "Test malformed3 $"),
            (
                "Test malformed4 This is $foo(a=b and $bar(",
                "Test malformed4 This is $foo(a=b and $bar(",
            ),
            (
                "Test malformed5 This is $foo(a=b, and $repl()",
                "Test malformed5 This is $foo(a=b, and rr",
            ),
            ("Test nonstr 4x2 = $double(4)", "Test nonstr 4x2 = 8"),
            ("Test nonstr 4x2 = $double(foo)", "Test nonstr 4x2 = N/A"),
            ("Test clr $clr(r, This is a red string!)", "Test clr |rThis is a red string!|n"),
            ("Test eval1 $eval(21 + 21 - 10)", "Test eval1 32"),
            ("Test eval2 $eval((21 + 21) / 2)", "Test eval2 21.0"),
            ("Test eval3 $eval(\"'21' + 'foo' + 'bar'\")", "Test eval3 21foobar"),
            (r"Test eval4 $eval('21' + '$repl()' + \"\" + str(10 // 2))", "Test eval4 21rr5"),
            (
                r"Test eval5 $eval(\'21\' + \'\$repl()\' + \'\' + str(10 // 2))",
                "Test eval5 21$repl()5",
            ),
            ("Test eval6 $eval(\"'$repl(a)' + '$repl(b)'\")", "Test eval6 rarrbr"),
            ("Test type1 $typ([1,2,3,4])", "Test type1 <class 'list'>"),
            ("Test type2 $typ((1,2,3,4))", "Test type2 <class 'tuple'>"),
            ("Test type3 $typ({1,2,3,4})", "Test type3 <class 'set'>"),
            ("Test type4 $typ({1:2,3:4})", "Test type4 <class 'dict'>"),
            ("Test type5 $typ(1), $typ(1.0)", "Test type5 <class 'int'>, <class 'float'>"),
            (
                "Test type6 $typ(\"'1'\"), $typ('\"1.0\"')",
                "Test type6 <class 'str'>, <class 'str'>",
            ),
            ("Test add1 $add(1, 2)", "Test add1 3"),
            ("Test add2 $add([1,2,3,4], [5,6])", "Test add2 [1, 2, 3, 4, 5, 6]"),
            ("Test literal1 $sum($lit([1,2,3,4,5,6]))", "Test literal1 21"),
            ("Test literal2 $typ($lit(1))", "Test literal2 <class 'int'>"),
            ("Test literal3 $typ($lit(1)aaa)", "Test literal3 <class 'str'>"),
            ("Test literal4 $typ(aaa$lit(1))", "Test literal4 <class 'str'>"),
            ("Test spider's thread", "Test spider's thread"),
        ]
    )
    def test_parse(self, string, expected):
        """
        Test parsing of string.

        """
        # t0 = time.time()
        # from evennia import set_trace;set_trace()
        ret = self.parser.parse(string, raise_errors=True)
        # t1 = time.time()
        # print(f"time: {(t1-t0)*1000} ms")
        self.assertEqual(expected, ret)

    @parameterized.expand(
        (
            "Test malformed This is $dummy(a, b) and $bar(",
            "Test $funcNotFound()",
        )
    )
    def test_parse_raise_unparseable(self, unparseable):
        """
        Make sure error is raised if told to do so.

        """
        with self.assertRaises(funcparser.ParsingError):
            self.parser.parse(unparseable, raise_errors=True)

    @parameterized.expand(
        [
            # max_nest, cause error for 4 nested funcs?
            (0, False),
            (1, False),
            (2, False),
            (3, False),
            (4, True),
            (5, True),
            (6, True),
        ]
    )
    def test_parse_max_nesting(self, max_nest, ok):
        """
        Make sure it is an error if the max nesting value is reached. We test
        four nested functions against differnt MAX_NESTING values.

        TODO: Does this make sense? When it sees the first function, len(callstack)
        is 0. It doesn't raise until the stack length is greater than the
        _MAX_NESTING value, which means you can nest 4 values with a value of
        2, as demonstrated by this test.
        """
        string = "$add(1, $add(1, $add(1, $eval(42))))"

        with patch("evennia.utils.funcparser._MAX_NESTING", max_nest):
            if ok:
                ret = self.parser.parse(string, raise_errors=True)
                self.assertEqual(ret, "45")
            else:
                with self.assertRaises(funcparser.ParsingError):
                    self.parser.parse(string, raise_errors=True)

    def test_parse_underlying_exception(self):
        string = "test $add(1, 1) $raise()"
        ret = self.parser.parse(string)

        # TODO: Does this return value actually make sense?
        # It completed the first function call.
        self.assertEqual("test 2 $raise()", ret)

        with self.assertRaises(RuntimeError):
            self.parser.parse(string, raise_errors=True)

    def test_parse_strip(self):
        """
        Test the parser's strip functionality.

        """
        string = "Test $foo(a,b, $bar()) and $repl($eval(3+2)) things"
        ret = self.parser.parse(string, strip=True)
        self.assertEqual("Test  and  things", ret)

    def test_parse_whitespace_preserved(self):
        string = "The answer is $foobar(1, x)"  # not found, so should be preserved
        ret = self.parser.parse(string)
        self.assertEqual("The answer is $foobar(1, x)", ret)

        string = 'The $pass(testing,  bar= $dum(b = "test2" , a), ) $pass('
        ret = self.parser.parse(string)
        self.assertEqual("The ('testing',){'bar': '$dum(b = \"test2\" , a)'} $pass(", ret)

    def test_parse_escape(self):
        """
        Test the parser's escape functionality.

        """
        string = "Test $foo(a) and $bar() and $rep(c) things"
        ret = self.parser.parse(string, escape=True)
        self.assertEqual("Test \$foo(a) and \$bar() and \$rep(c) things", ret)

    def test_parse_lit(self):
        """
        Get non-strings back from parsing.

        """
        string = "$lit(123)"

        # normal parse
        ret = self.parser.parse(string)
        self.assertEqual("123", ret)
        self.assertTrue(isinstance(ret, str))

        # parse lit
        ret = self.parser.parse_to_any(string)
        self.assertEqual(123, ret)
        self.assertTrue(isinstance(ret, int))

        ret = self.parser.parse_to_any("$lit([1,2,3,4])")
        self.assertEqual([1, 2, 3, 4], ret)
        self.assertTrue(isinstance(ret, list))

        ret = self.parser.parse_to_any("$lit(\"''\")")
        self.assertEqual("", ret)
        self.assertTrue(isinstance(ret, str))

        ret = self.parser.parse_to_any(r"$lit(\'\')")
        self.assertEqual("", ret)
        self.assertTrue(isinstance(ret, str))

        # mixing a literal with other chars always make a string
        ret = self.parser.parse_to_any(string + "aa")
        self.assertEqual("123aa", ret)
        self.assertTrue(isinstance(ret, str))

        ret = self.parser.parse_to_any("test")
        self.assertEqual("test", ret)
        self.assertTrue(isinstance(ret, str))

    def test_kwargs_overrides(self):
        """
        Test so default kwargs are added and overridden properly

        """
        # default kwargs passed on initializations
        parser = funcparser.FuncParser(_test_callables, test="foo")
        ret = parser.parse("This is a $foo() string")
        self.assertEqual("This is a _test(test=foo) string", ret)

        # override in the string itself

        ret = parser.parse("This is a $foo(test=bar,foo=moo) string")
        self.assertEqual("This is a _test(test=bar, foo=moo) string", ret)

        # parser kwargs override the other types

        ret = parser.parse("This is a $foo(test=bar,foo=moo) string", test="override", foo="bar")
        self.assertEqual("This is a _test(test=override, foo=bar) string", ret)

        # non-overridden kwargs shine through

        ret = parser.parse("This is a $foo(foo=moo) string", foo="bar")
        self.assertEqual("This is a _test(test=foo, foo=bar) string", ret)


class _DummyObj:
    def __init__(self, name):
        self.name = name

    def get_display_name(self, looker=None):
        return self.name


class TestDefaultCallables(TestCase):
    """
    Test default callables.

    """

    def setUp(self):
        from django.conf import settings

        self.parser = funcparser.FuncParser(
            {**funcparser.FUNCPARSER_CALLABLES, **funcparser.ACTOR_STANCE_CALLABLES}
        )

        self.obj1 = _DummyObj("Char1")
        self.obj2 = _DummyObj("Char2")

    @parameterized.expand(
        [
            ("Test py1 $eval('')", "Test py1 "),
        ]
    )
    def test_callable(self, string, expected):
        """
        Test callables with various input strings

        """
        ret = self.parser.parse(string, raise_errors=True)
        self.assertEqual(expected, ret)

    @parameterized.expand(
        [
            ("$You() $conj(smile) at him.", "You smile at him.", "Char1 smiles at him."),
            ("$You() $conj(smile) at $You(char1).", "You smile at You.", "Char1 smiles at Char1."),
            ("$You() $conj(smile) at $You(char2).", "You smile at Char2.", "Char1 smiles at You."),
            (
                "$You(char2) $conj(smile) at $you(char1).",
                "Char2 smile at you.",
                "You smiles at Char1.",
            ),
            (
                "$You() $conj(smile) to $pron(yourself,m).",
                "You smile to yourself.",
                "Char1 smiles to himself.",
            ),
            (
                "$You() $conj(smile) to $pron(herself).",
                "You smile to yourself.",
                "Char1 smiles to herself.",
            ),  # reverse reference
            (
                "$Your() smile is the greatest ever.",
                "Your smile is the greatest ever.",
                "Char1's smile is the greatest ever.",
            ),
        ]
    )
    def test_conjugate(self, string, expected_you, expected_them):
        """
        Test the $conj(), $you() and $pron callables with various input strings.
        """
        mapping = {"char1": self.obj1, "char2": self.obj2}
        ret = self.parser.parse(
            string, caller=self.obj1, receiver=self.obj1, mapping=mapping, raise_errors=True
        )
        self.assertEqual(expected_you, ret)
        ret = self.parser.parse(
            string, caller=self.obj1, receiver=self.obj2, mapping=mapping, raise_errors=True
        )
        self.assertEqual(expected_them, ret)

    def test_conjugate__non_existing_verb(self):
        """
        Test $conj() but with a nonsense verb. It should just return the verb as-is.

        """
        string = "$You() $conj(squanch) greatly."

        mapping = {"char1": self.obj1, "char2": self.obj2}
        ret = self.parser.parse(
            string, caller=self.obj1, receiver=self.obj1, mapping=mapping, raise_errors=True
        )
        self.assertEqual("You squanch greatly.", ret)
        ret = self.parser.parse(
            string, caller=self.obj1, receiver=self.obj2, mapping=mapping, raise_errors=True
        )
        self.assertEqual("Char1 squanchs greatly.", ret)

    def test_conjugate_missing_args(self):
        string = "You $conj(smile)"
        with self.assertRaises(funcparser.ParsingError):
            self.parser.parse(string, raise_errors=True)

    @parameterized.expand(
        [
            ("male", "Char1 smiles at himself"),
            ("female", "Char1 smiles at herself"),
            ("neutral", "Char1 smiles at itself"),
            ("plural", "Char1 smiles at themselves"),
        ]
    )
    def test_pronoun_gender(self, gender, expected):
        string = "Char1 smiles at $pron(yourself)"

        self.obj1.gender = gender
        ret = self.parser.parse(string, caller=self.obj1, raise_errors=True)
        self.assertEqual(expected, ret)

        self.obj1.gender = lambda: gender
        ret = self.parser.parse(string, caller=self.obj1, raise_errors=True)
        self.assertEqual(expected, ret)

    def test_pronoun_viewpoint(self):
        string = "Char1 smiles at $pron(I)"

        ret = self.parser.parse(string, caller=self.obj1, viewpoint="op", raise_errors=True)
        self.assertEqual("Char1 smiles at it", ret)

    def test_pronoun_capitalize(self):
        string = "Char1 smiles at $pron(I)"

        ret = self.parser.parse(string, caller=self.obj1, capitalize=True, raise_errors=True)
        self.assertEqual("Char1 smiles at It", ret)

        string = "Char1 smiles at $Pron(I)"
        ret = self.parser.parse(string, caller=self.obj1, capitalize=True, raise_errors=True)
        self.assertEqual("Char1 smiles at It", ret)

    @parameterized.expand(
        [
            ("Test $pad(Hello, 20, c, -) there", "Test -------Hello-------- there"),
            (
                "Test $pad(Hello, width=20, align=c, fillchar=-) there",
                "Test -------Hello-------- there",
            ),
            ("Test $crop(This is a long test, 12)", "Test This is[...]"),
            ("Some $space(10) here", "Some            here"),
            ("Some $clr(b, blue color) now", "Some |bblue color|n now"),
            ("Some $add(1, 2) things", "Some 3 things"),
            ("Some $sub(10, 2) things", "Some 8 things"),
            ("Some $mult(3, 2) things", "Some 6 things"),
            ("Some $div(6, 2) things", "Some 3.0 things"),
            ("Some $toint(6) things", "Some 6 things"),
            ("Some $toint(3 + 3) things", "Some 6 things"),
            ("Some $ljust(Hello, 30)", "Some Hello                         "),
            ("Some $rjust(Hello, 30)", "Some                          Hello"),
            ("Some $rjust(Hello, width=30)", "Some                          Hello"),
            ("Some $cjust(Hello, 30)", "Some             Hello             "),
            (
                "There $pluralize(is, 1, are) one $pluralize(goose, 1, geese) here.",
                "There is one goose here.",
            ),
            (
                "There $pluralize(is, 2, are) two $pluralize(goose, 2, geese) here.",
                "There are two geese here.",
            ),
            (
                "There is $int2str(1) murderer, but $int2str(12) suspects.",
                "There is one murderer, but twelve suspects.",
            ),
            ("There is $an(thing) here", "There is a thing here"),
            ("Some $eval(\"'-'*20\")Hello", "Some --------------------Hello"),
            ('$crop("spider\'s silk", 5)', "spide"),
            ("$crop(spider's silk, 5)", "spide"),
            ("$an(apple)", "an apple"),
            ("$round(2.9) apples", "3.0 apples"),
            ("$round(2.967, 1) apples", "3.0 apples"),
            # Degenerate cases
            ("$int2str() apples", " apples"),
            ("$int2str(x) apples", "x apples"),
            ("$int2str(1 + 1) apples", "1 + 1 apples"),
            ("$int2str(13) apples", "13 apples"),
            ("$toint([1, 2, 3]) apples", "[1, 2, 3] apples"),
            ("$an() foo bar", " foo bar"),
            ("$add(1) apple", " apple"),
            ("$add(1, [1, 2]) apples", " apples"),
            ("$round() apples", " apples"),
            ("$choice() apple", " apple"),
            ("A $pad() apple", "A  apple"),
            ("A $pad(tasty, 13, x, -) apple", "A ----tasty---- apple"),
            ("A $crop() apple", "A  apple"),
            ("A $space() apple", "A  apple"),
            ("A $justify() apple", "A  apple"),
            ("A $clr() apple", "A  apple"),
            ("A $clr(red) apple", "A red apple"),
            ("10 $pluralize()", "10 "),
            ("10 $pluralize(apple, 10)", "10 apples"),
            ("1 $pluralize(apple)", "1 apple"),
            ("You $conj()", "You "),
            ("$pron() smiles", " smiles"),
        ]
    )
    def test_other_callables(self, string, expected):
        """
        Test default callables.

        """
        ret = self.parser.parse(string, raise_errors=True)
        self.assertEqual(expected, ret)

    def test_random(self):
        """
        Test random callable, with ranges of expected values.
        """
        string = "$random(1,10)"
        for i in range(100):
            ret = self.parser.parse_to_any(string, raise_errors=True)
            self.assertTrue(1 <= ret <= 10)

        string = "$random()"
        for i in range(100):
            ret = self.parser.parse_to_any(string, raise_errors=True)
            self.assertTrue(0 <= ret <= 1)

        string = "$random(2)"
        for i in range(100):
            ret = self.parser.parse_to_any(string, raise_errors=True)
            self.assertTrue(0 <= ret <= 2)

        string = "$random(1.0, 3.0)"
        for i in range(100):
            ret = self.parser.parse_to_any(string, raise_errors=True)
            self.assertTrue(isinstance(ret, float))
            self.assertTrue(1.0 <= ret <= 3.0)

        string = "$random([1,2]) apples"
        ret = self.parser.parse_to_any(string)
        self.assertEqual(" apples", ret)
        with self.assertRaises(TypeError):
            ret = self.parser.parse_to_any(string, raise_errors=True)

    # @unittest.skip("underlying function seems broken")
    def test_choice(self):
        """
        Test choice callable, where output could be either choice.
        """
        string = "$choice(red, green) apple"
        ret = self.parser.parse(string)
        self.assertIn(ret, ("red apple", "green apple"))

        string = "$choice([red, green]) apple"
        ret = self.parser.parse(string)
        self.assertIn(ret, ("red apple", "green apple"))

        string = "$choice(['red', 'green']) apple"
        ret = self.parser.parse(string)
        self.assertIn(ret, ("red apple", "green apple"))

        string = "$choice([1, 2])"
        ret = self.parser.parse(string)
        self.assertIn(ret, ("1", "2"))
        ret = self.parser.parse_to_any(string)
        self.assertIn(ret, (1, 2))

        string = "$choice(1, 2)"
        ret = self.parser.parse(string)
        self.assertIn(ret, ("1", "2"))
        ret = self.parser.parse_to_any(string)
        self.assertIn(ret, (1, 2))

    def test_choice_quotes(self):
        """
        Test choice, but also commas embedded.
        """

        string = "$choice(spider's, devil's, mummy's, zombie's)"
        ret = self.parser.parse(string)
        self.assertIn(ret, ("spider's", "devil's", "mummy's", "zombie's"))

        string = '$choice("Tiamat, queen of dragons", "Dracula, lord of the night")'
        ret = self.parser.parse(string)
        self.assertIn(ret, ("Tiamat, queen of dragons", "Dracula, lord of the night"))

        # single quotes are ignored, so this becomes many entries
        string = "$choice('Tiamat, queen of dragons', 'Dracula, lord of the night')"
        ret = self.parser.parse(string)
        self.assertIn(ret, ("'Tiamat", "queen of dragons'", "'Dracula", "lord of the night'"))

    def test_randint(self):
        string = "$randint(1.0, 3.0)"
        ret = self.parser.parse_to_any(string, raise_errors=True)
        self.assertTrue(isinstance(ret, int))
        self.assertTrue(1.0 <= ret <= 3.0)

    def test_nofunc(self):
        self.assertEqual(
            self.parser.parse("as$382ewrw w we w werw,|44943}"),
            "as$382ewrw w we w werw,|44943}",
        )

    def test_incomplete(self):
        self.assertEqual(
            self.parser.parse("testing $blah{without an ending."),
            "testing $blah{without an ending.",
        )

    def test_single_func(self):
        self.assertEqual(
            self.parser.parse("this is a test with $pad(centered, 20) text in it."),
            "this is a test with       centered       text in it.",
        )

    def test_nested(self):
        self.assertEqual(
            self.parser.parse(
                "this $crop(is a test with $pad(padded, 20) text in $pad(pad2, 10) a crop, 80)"
            ),
            "this is a test with        padded        text in    pad2    a crop",
        )

    def test_escaped(self):
        raw_str = (
            'this should be $pad("""escaped,""" and """instead,""" cropped $crop(with a long,5)'
            " text., 80)"
        )
        expected = (
            "this should be                    escaped, and instead, cropped with  text.           "
            "         "
        )
        result = self.parser.parse(raw_str)
        self.assertEqual(
            result,
            expected,
        )


class TestCallableSearch(test_resources.BaseEvenniaTest):
    """
    Test the $search(query) callable

    """

    def setUp(self):
        super().setUp()
        self.parser = funcparser.FuncParser(funcparser.SEARCHING_CALLABLES)

    def test_search_obj(self):
        """
        Test searching for an object

        """
        string = "$search(Char)"
        expected = self.char1

        ret = self.parser.parse(string, caller=self.char1, return_str=False, raise_errors=True)
        self.assertEqual(expected, ret)

    def test_search_account(self):
        """
        Test searching for an account

        """
        string = "$search(TestAccount, type=account)"
        expected = self.account
        self.account.locks.add("control:id(%s)" % self.char1.dbref)

        ret = self.parser.parse(string, caller=self.char1, return_str=False, raise_errors=True)
        self.assertEqual(expected, ret)

    def test_search_script(self):
        """
        Test searching for a script

        """
        string = "$search(Script, type=script)"
        expected = self.script
        self.script.locks.add("control:id(%s)" % self.char1.dbref)

        ret = self.parser.parse(string, caller=self.char1, return_str=False, raise_errors=True)
        self.assertEqual(expected, ret)

    def test_search_obj_embedded(self):
        """
        Test searching for an object - embedded in str

        """
        string = "This is $search(Char) the guy."
        expected = "This is " + str(self.char1) + " the guy."

        ret = self.parser.parse(string, caller=self.char1, return_str=False, raise_errors=True)
        self.assertEqual(expected, ret)

    def test_search_tag(self):
        """
        Test searching for a tag
        """
        self.char1.tags.add("foo")

        string = "This is $search(foo, type=tag)"
        expected = "This is %s" % str(self.char1)

        ret = self.parser.parse(string, caller=self.char1, return_str=False, raise_errors=True)
        self.assertEqual(expected, ret)

    def test_search_not_found(self):
        string = "$search(foo)"
        with self.assertRaises(funcparser.ParsingError):
            self.parser.parse(string, caller=self.char1, return_str=False, raise_errors=True)

        ret = self.parser.parse(string, caller=self.char1, return_str=False, raise_errors=False)
        self.assertEqual("$search(foo)", ret)

        ret = self.parser.parse_to_any(
            string, caller=self.char1, return_list=True, raise_errors=False
        )
        self.assertEqual([], ret)

    def test_search_multiple_results_no_list(self):
        """
        Test exception when search returns multiple results but list is not requested
        """
        string = "$search(BaseObject)"
        with self.assertRaises(funcparser.ParsingError):
            self.parser.parse(string, caller=self.char1, return_str=False, raise_errors=True)

    def test_search_no_access(self):
        string = "Go to $search(Room)"
        with self.assertRaises(funcparser.ParsingError):
            self.parser.parse(string, caller=self.char2, return_list=True, raise_errors=True)

    def test_search_no_caller(self):
        string = "$search(Char)"
        with self.assertRaises(funcparser.ParsingError):
            self.parser.parse(string, caller=None, raise_errors=True)

    def test_search_no_args(self):
        string = "$search()"
        ret = self.parser.parse(string, caller=self.char1, return_list=False, raise_errors=True)
        self.assertEqual("None", ret)

        ret = self.parser.parse(string, caller=self.char1, return_list=True, raise_errors=True)
        self.assertEqual("[]", ret)

    def test_search_nested__issue2902(self):
        """
        Search for objects by-tag, check that the result is a valid object

        """
        # we
        parser = funcparser.FuncParser(
            {**funcparser.SEARCHING_CALLABLES, **funcparser.FUNCPARSER_CALLABLES}
        )

        # set up search targets
        self.obj1.tags.add("beach", category="zone")
        self.obj2.tags.add("beach", category="zone")

        # first a plain search
        string = "$objlist(beach,category=zone,type=tag)"
        ret = parser.parse_to_any(string, caller=self.char1, raise_errors=True)

        self.assertEqual(ret, [self.obj1, self.obj2])

        # get random result from the possible matches
        string = "$choice($objlist(beach,category=zone,type=tag))"
        ret = parser.parse_to_any(string, caller=self.char1, raise_errors=True)

        self.assertIn(ret, [self.obj1, self.obj2])

        # test wrapping in $obj(), should just pass object through
        string = "$obj($choice($objlist(beach,category=zone,type=tag)))"
        ret = parser.parse_to_any(string, caller=self.char1, raise_errors=True)

        self.assertIn(ret, [self.obj1, self.obj2])



"""
EvAdventure NPCs. This includes both friends and enemies, only separated by their AI.

"""
from random import choice

from evennia import DefaultCharacter
from evennia.typeclasses.attributes import AttributeProperty
from evennia.typeclasses.tags import TagProperty
from evennia.utils.evmenu import EvMenu
from evennia.utils.utils import make_iter

from .characters import LivingMixin
from .enums import Ability, WieldLocation
from .objects import get_bare_hands
from .rules import dice


class EvAdventureNPC(LivingMixin, DefaultCharacter):
    """
    This is the base class for all non-player entities, including monsters. These
    generally don't advance in level but uses a simplified, abstract measure of how
    dangerous or competent they are - the 'hit dice' (HD).

    HD indicates how much health they have and how hard they hit. In _Knave_, HD also
    defaults to being the bonus for all abilities. HP is 4 x Hit die (this can then be
    customized per-entity of course).

    Morale is set explicitly per-NPC, usually between 7 and 9.

    Monsters don't use equipment in the way PCs do, instead they have a fixed armor
    value, and their Abilities are dynamically generated from the HD (hit_dice).

    If wanting monsters or NPCs that can level and work the same as PCs, base them off the
    EvAdventureCharacter class instead.

    The weapon of the npc is stored as an Attribute instead of implementing a full
    inventory/equipment system. This means that the normal inventory can be used for
    non-combat purposes (or for loot to get when killing an enemy).

    """

    is_pc = False

    hit_dice = AttributeProperty(default=1, autocreate=False)
    armor = AttributeProperty(default=1, autocreate=False)  # +10 to get armor defense
    morale = AttributeProperty(default=9, autocreate=False)
    hp_multiplier = AttributeProperty(default=4, autocreate=False)  # 4 default in Knave
    hp = AttributeProperty(default=None, autocreate=False)  # internal tracking, use .hp property
    allegiance = AttributeProperty(default=Ability.ALLEGIANCE_HOSTILE, autocreate=False)

    is_idle = AttributeProperty(default=False, autocreate=False)

    weapon = AttributeProperty(default=get_bare_hands, autocreate=False)  # instead of inventory
    coins = AttributeProperty(default=1, autocreate=False)  # coin loot

    # if this npc is attacked, everyone with the same tag in the current location will also be
    # pulled into combat.
    group = TagProperty("npcs")

    @property
    def strength(self):
        return self.hit_dice

    @property
    def dexterity(self):
        return self.hit_dice

    @property
    def constitution(self):
        return self.hit_dice

    @property
    def intelligence(self):
        return self.hit_dice

    @property
    def wisdom(self):
        return self.hit_dice

    @property
    def charisma(self):
        return self.hit_dice

    @property
    def hp_max(self):
        return self.hit_dice * self.hp_multiplier

    def at_object_creation(self):
        """
        Start with max health.

        """
        self.hp = self.hp_max
        self.tags.add("npcs", category="group")

    def at_attacked(self, attacker, **kwargs):
        """
        Called when being attacked and combat starts.

        """
        pass

    def ai_next_action(self, **kwargs):
        """
        The combat engine should ask this method in order to
        get the next action the npc should perform in combat.

        """
        pass


class EvAdventureTalkativeNPC(EvAdventureNPC):
    """
    Talkative NPCs can be addressed by `talk [to] <npc>`. This opens a chat menu with
    communication options. The menu is created with the npc and we override the .create
    to allow passing in the menu nodes.

    """

    menudata = AttributeProperty(dict(), autocreate=False)
    menu_kwargs = AttributeProperty(dict(), autocreate=False)
    # text shown when greeting at the start of a conversation. If this is an
    # iterable, a random reply will be chosen by the menu
    hi_text = AttributeProperty("Hi!", autocreate=False)

    def at_damage(self, damage, attacker=None):
        """
        Talkative NPCs are generally immortal (we don't deduct HP here by default)."

        """
        attacker.msg(f'{self.key} dodges the damage and shouts "|wHey! What are you doing?|n"')

    @classmethod
    def create(cls, key, account=None, **kwargs):
        """
        Overriding the creation of the NPC, allowing some extra `**kwargs`.

        Args:
            key (str): Name of the new object.
            account (Account, optional): Account to attribute this object to.

        Keyword Args:
            description (str): Brief description for this object (same as default Evennia)
            ip (str): IP address of creator (for object auditing) (same as default Evennia).
            menudata (dict or str): The `menudata` argument to `EvMenu`. This is either a dict of
                `{"nodename": <node_callable>,...}` or the python-path to a module containing
                such nodes (see EvMenu docs). This will be used to generate the chat menu
                chat menu for the character that talks to the NPC (which means the `at_talk` hook
                is called (by our custom `talk` command).
            menu_kwargs (dict): This will be passed as `**kwargs` into `EvMenu` when it
                is created. Make sure this dict can be pickled to an Attribute.

        Returns:
            tuple: `(new_character, errors)`. On error, the `new_character` is `None` and
            `errors` is a `list` of error strings (an empty list otherwise).


        """
        menudata = kwargs.pop("menudata", None)
        menu_kwargs = kwargs.pop("menu_kwargs", {})

        # since this is a @classmethod we can't use super() here
        new_object, errors = EvAdventureNPC.create(
            key, account=account, attributes=(("menudata", menudata), ("menu_kwargs", menu_kwargs))
        )

        return new_object, errors

    def at_talk(self, talker, startnode="node_start", session=None, **kwargs):
        """
        Called by the `talk` command when another entity addresses us.

        Args:
            talker (Object): The one talking to us.
            startnode (str, optional): Allows to start in a different location in the menu tree.
                The given node must exist in the tree.
            session (Session, optional): The talker's current session, allows for routing
                correctly in multi-session modes.
            **kwargs: This will be passed into the `EvMenu` creation and appended and `menu_kwargs`
                given to the NPC at creation.

        Notes:
            We pass `npc=self` into the EvMenu for easy back-reference. This will appear in the
            `**kwargs` of the start node.

        """
        menu_kwargs = {**self.menu_kwargs, **kwargs}
        EvMenu(talker, self.menudata, startnode=startnode, session=session, npc=self, **menu_kwargs)


def node_start(caller, raw_string, **kwargs):
    """
    This is the intended start menu node for the Talkative NPC interface. It will
    use on-npc Attributes to build its message and will also pick its options
    based on nodes named `node_start_*` are available in the node tree.

    """
    # we presume a back-reference to the npc this is added when the menu is created
    npc = kwargs["npc"]

    # grab a (possibly random) welcome text
    text = choice(make_iter(npc.hi_text))

    # determine options based on `node_start_*` nodes available
    toplevel_node_keys = [
        node_key for node_key in caller.ndb._evmenu._menutree if node_key.startswith("node_start_")
    ]
    options = []
    for node_key in toplevel_node_keys:
        option_name = node_key[11:].replace("_", " ").capitalized()

        # we let the menu number the choices, so we don't use key here
        options.append({"desc": option_name, "goto": node_key})

    return text, options


class EvAdventureQuestGiver(EvAdventureTalkativeNPC):
    """
    An NPC that acts as a dispenser of quests.

    """


class EvAdventureShopKeeper(EvAdventureTalkativeNPC):
    """
    ShopKeeper NPC.

    """

    # how much extra the shopkeeper adds on top of the item cost
    upsell_factor = AttributeProperty(1.0, autocreate=False)
    # how much of the raw cost the shopkeep is willing to pay when buying from character
    miser_factor = AttributeProperty(0.5, autocreate=False)
    # prototypes of common wares
    common_ware_prototypes = AttributeProperty([], autocreate=False)

    def at_damage(self, damage, attacker=None):
        """
        Immortal - we don't deduct any damage here.

        """
        attacker.msg(
            f"{self.key} brushes off the hit and shouts "
            '"|wHey! This is not the way to get a discount!|n"'
        )


class EvAdventureMob(EvAdventureNPC):
    """
    Mob (mobile) NPC; this is usually an enemy.

    """

    # chance (%) that this enemy will loot you when defeating you
    loot_chance = AttributeProperty(75, autocreate=False)

    def ai_next_action(self, **kwargs):
        """
        Called to get the next action in combat.

        Args:
            combathandler (EvAdventureCombatHandler): The currently active combathandler.

        Returns:
            tuple: A tuple `(str, tuple, dict)`, being the `action_key`, and the `*args` and
            `**kwargs` for that action. The action-key is that of a CombatAction available to the
            combatant in the current combat handler.

        """
        from .combat import CombatActionAttack, CombatActionDoNothing

        if self.is_idle:
            # mob just stands around
            return CombatActionDoNothing.key, (), {}

        target = choice(combathandler.get_enemy_targets(self))

        # simply randomly decide what action to take
        action = choice(
            (
                CombatActionAttack,
                CombatActionDoNothing,
            )
        )
        return action.key, (target,), {}

    def at_defeat(self):
        """
        Mobs die right away when defeated, no death-table rolls.

        """
        self.at_death()

    def at_do_loot(self, looted):
        """
        Called when mob gets to loot a PC.

        """
        if dice.roll("1d100") > self.loot_chance:
            # don't loot
            return

        if looted.coins:
            # looter prefer coins
            loot = dice.roll("1d20")
            if looted.coins < loot:
                self.location.msg_location(
                    "$You(looter) loots $You() for all coin!",
                    from_obj=looted,
                    mapping={"looter": self},
                )
            else:
                self.location.msg_location(
                    "$You(looter) loots $You() for |y{loot}|n coins!",
                    from_obj=looted,
                    mapping={"looter": self},
                )
        elif hasattr(looted, "equipment"):
            # go through backpack, first usable, then wieldable, wearable items
            # and finally stuff wielded
            stealable = looted.equipment.get_usable_objects_from_backpack()
            if not stealable:
                stealable = looted.equipment.get_wieldable_objects_from_backpack()
            if not stealable:
                stealable = looted.equipment.get_wearable_objects_from_backpack()
            if not stealable:
                stealable = [looted.equipment.slots[WieldLocation.SHIELD_HAND]]
            if not stealable:
                stealable = [looted.equipment.slots[WieldLocation.HEAD]]
            if not stealable:
                stealable = [looted.equipment.slots[WieldLocation.ARMOR]]
            if not stealable:
                stealable = [looted.equipment.slots[WieldLocation.WEAPON_HAND]]
            if not stealable:
                stealable = [looted.equipment.slots[WieldLocation.TWO_HANDS]]

            stolen = looted.equipment.remove(choice(stealable))
            stolen.location = self

            self.location.msg_location(
                "$You(looter) steals {stolen.key} from $You()!",
                from_obj=looted,
                mapping={"looter": self},
            )



"""
This is the starting point when a user enters a url in their web browser.

The urls is matched (by regex) and mapped to a 'view' - a Python function or
callable class that in turn (usually) makes use of a 'template' (a html file
with slots that can be replaced by dynamic content) in order to render a HTML
page to show the user.

This file includes the urls in website, webclient and admin. To override you
should modify urls.py in those sub directories.

Search the Django documentation for "URL dispatcher" for more help.

"""
from django.urls import include, path

# default evennia patterns
from evennia.web.urls import urlpatterns as evennia_default_urlpatterns

# add patterns
urlpatterns = [
    # website
    path("", include("web.website.urls")),
    # webclient
    path("webclient/", include("web.webclient.urls")),
    # web admin
    path("admin/", include("web.admin.urls")),
    # add any extra urls here:
    # path("mypath/", include("path.to.my.urls.file")),
]

# 'urlpatterns' must be named such for Django to find it.
urlpatterns = urlpatterns + evennia_default_urlpatterns



"""
Language and whisper obfuscation system

Evennia contrib - Griatch 2015

This module is intented to be used with an emoting system (such as
contrib/rpsystem.py). It offers the ability to obfuscate spoken words
in the game in various ways:

- Language: The language functionality defines a pseudo-language map
    to any number of languages.  The string will be obfuscated depending
    on a scaling that (most likely) will be input as a weighted average of
    the language skill of the speaker and listener.
- Whisper: The whisper functionality will gradually "fade out" a
    whisper along as scale 0-1, where the fading is based on gradually
    removing sections of the whisper that is (supposedly) easier to
    overhear (for example "s" sounds tend to be audible even when no other
    meaning can be determined).

## Usage

```python
from evennia.contrib import rplanguage

# need to be done once, here we create the "default" lang
rplanguage.add_language()

say = "This is me talking."
whisper = "This is me whispering.

print rplanguage.obfuscate_language(say, level=0.0)
<<< "This is me talking."
print rplanguage.obfuscate_language(say, level=0.5)
<<< "This is me byngyry."
print rplanguage.obfuscate_language(say, level=1.0)
<<< "Daly ly sy byngyry."

result = rplanguage.obfuscate_whisper(whisper, level=0.0)
<<< "This is me whispering"
result = rplanguage.obfuscate_whisper(whisper, level=0.2)
<<< "This is m- whisp-ring"
result = rplanguage.obfuscate_whisper(whisper, level=0.5)
<<< "---s -s -- ---s------"
result = rplanguage.obfuscate_whisper(whisper, level=0.7)
<<< "---- -- -- ----------"
result = rplanguage.obfuscate_whisper(whisper, level=1.0)
<<< "..."

```

## Custom languages

To set up new languages, you need to run `add_language()`
helper function in this module. The arguments of this function (see below)
are used to store the new language in the database (in the LanguageHandler,
which is a type of Script).

If you want to remember the language definitions, you could put them all
in a module along with the `add_language` call as a quick way to
rebuild the language on a db reset:

```python
# a stand-alone module somewhere under mygame. Just import this
# once to automatically add the language!

from evennia.contrib.rpg.rpsystem import rplanguage
grammar = (...)
vowels = "eaouy"
# etc

rplanguage.add_language(grammar=grammar, vowels=vowels, ...)
```

The variables of `add_language` allows you to customize the "feel" of
the semi-random language you are creating. Especially
the `word_length_variance` helps vary the length of translated
words compared to the original. You can also add your own
dictionary and "fix" random words for a list of input words.

## Example

Below is an example module creating "elvish", using "rounder" vowels and sounds:

```python
# vowel/consonant grammar possibilities
grammar = ("v vv vvc vcc vvcc cvvc vccv vvccv vcvccv vcvcvcc vvccvvcc "
           "vcvvccvvc cvcvvcvvcc vcvcvvccvcvv")

# all not in this group is considered a consonant
vowels = "eaoiuy"

# you need a representative of all of the minimal grammars here, so if a
# grammar v exists, there must be atleast one phoneme available with only
# one vowel in it
phonemes = ("oi oh ee ae aa eh ah ao aw ay er ey ow ia ih iy "
            "oy ua uh uw y p b t d f v t dh s z sh zh ch jh k "
            "ng g m n l r w")

# how much the translation varies in length compared to the original. 0 is
# smallest, higher values give ever bigger randomness (including removing
# short words entirely)
word_length_variance = 1

# if a proper noun (word starting with capitalized letter) should be
# translated or not. If not (default) it means e.g. names will remain
# unchanged across languages.
noun_translate = False

# all proper nouns (words starting with a capital letter not at the beginning
# of a sentence) can have either a postfix or -prefix added at all times
noun_postfix = "'la"

# words in dict will always be translated this way. The 'auto_translations'
# is instead a list or filename to file with words to use to help build a
# bigger dictionary by creating random translations of each word in the
# list *once* and saving the result for subsequent use.
manual_translations = {"the":"y'e", "we":"uyi", "she":"semi", "he":"emi",
                      "you": "do", 'me':'mi','i':'me', 'be':"hy'e", 'and':'y'}

rplanguage.add_language(key="elvish", phonemes=phonemes, grammar=grammar,
                         word_length_variance=word_length_variance,
                         noun_translate=noun_translate,
                         noun_postfix=noun_postfix, vowels=vowels,
                         manual_translations=manual_translations,
                         auto_translations="my_word_file.txt")

```

This will produce a decicively more "rounded" and "soft" language
than the default one. The few manual_translations also make sure
to make it at least look superficially "reasonable".

The `auto_translations` keyword is useful, this accepts either a
list or a path to a file of words (one per line) to automatically
create fixed translations for according to the grammatical rules.
This allows to quickly build a large corpus of translated words
that never change (if this is desired).

"""
import re
from collections import defaultdict
from random import choice, randint

from evennia import DefaultScript
from evennia.utils import logger

# ------------------------------------------------------------
#
# Obfuscate language
#
# ------------------------------------------------------------

# default language grammar
_PHONEMES = (
    "ea oh ae aa eh ah ao aw ai er ey ow ia ih iy oy ua uh uw a e i u y p b t d f v t dh "
    "s z sh zh ch jh k ng g m n l r w"
)
_VOWELS = "eaoiuy"
# these must be able to be constructed from phonemes (so for example,
# if you have v here, there must exist at least one single-character
# vowel phoneme defined above)
_GRAMMAR = "v cv vc cvv vcc vcv cvcc vccv cvccv cvcvcc cvccvcv vccvccvc cvcvccvv cvcvcvcvv"

_RE_FLAGS = re.MULTILINE + re.IGNORECASE + re.DOTALL + re.UNICODE
_RE_GRAMMAR = re.compile(r"vv|cc|v|c", _RE_FLAGS)
_RE_WORD = re.compile(r"\w+", _RE_FLAGS)
# superfluous chars, except ` ... `
_RE_EXTRA_CHARS = re.compile(r"\s+(?!... )(?=\W)|[,.?;](?!.. )(?=[,?;]|\s+[,.?;])", _RE_FLAGS)


class LanguageError(RuntimeError):
    pass


class LanguageExistsError(LanguageError):
    pass


class LanguageHandler(DefaultScript):
    """
    This is a storage class that should usually not be created on its
    own. It's automatically created by a call to `obfuscate_language`
    or `add_language` below.

    Languages are implemented as a "logical" pseudo- consistent language
    algorith here. The idea is that a language is built up from
    phonemes. These are joined together according to a "grammar" of
    possible phoneme- combinations and allowed characters. It may
    sound simplistic, but this allows to easily make
    "similar-sounding" languages. One can also custom-define a
    dictionary of some common words to give further consistency.
    Optionally, the system also allows an input list of common words
    to be loaded and given random translations. These will be stored
    to disk and will thus not change. This gives a decent "stability"
    of the language but if the goal is to obfuscate, this may allow
    players to eventually learn to understand the gist of a sentence
    even if their characters can not. Any number of languages can be
    created this way.

    This nonsense language will partially replace the actual spoken
    language when so desired (usually because the speaker/listener
    don't know the language well enough).

    """

    def at_script_creation(self):
        "Called when script is first started"
        self.key = "language_handler"
        self.persistent = True
        self.db.language_storage = {}

    def add(
        self,
        key="default",
        phonemes=_PHONEMES,
        grammar=_GRAMMAR,
        word_length_variance=0,
        noun_translate=False,
        noun_prefix="",
        noun_postfix="",
        vowels=_VOWELS,
        manual_translations=None,
        auto_translations=None,
        force=False,
    ):
        """
        Add a new language. Note that you generally only need to do
        this once per language and that adding an existing language
        will re-initialize all the random components to new permanent
        values.

        Args:
            key (str, optional): The name of the language. This
                will be used as an identifier for the language so it
                should be short and unique.
            phonemes (str, optional): Space-separated string of all allowed
                phonemes in this language. If either of the base phonemes
                (c, v, cc, vv) are present in the grammar, the phoneme list must
                at least include one example of each.
            grammar (str): All allowed consonant (c) and vowel (v) combinations
                allowed to build up words. Grammars are broken into the base phonemes
                (c, v, cc, vv) prioritizing the longer bases. So cvv would be a
                the c + vv (would allow for a word like 'die' whereas
                cvcvccc would be c+v+c+v+cc+c (a word like 'galosch').
            word_length_variance (real): The variation of length of words.
                0 means a minimal variance, higher variance may mean words
                have wildly varying length; this strongly affects how the
                language "looks".
            noun_translate (bool, optional): If a proper noun should be translated or
                not. By default they will not, allowing for e.g. the names of characters
                to be understandable. A 'noun' is identified as a capitalized word
                *not at the start of a sentence*. This simple metric means that names
                starting a sentence always will be translated (- but hey, maybe
                the fantasy language just never uses a noun at the beginning of
                sentences, who knows?)
            noun_prefix (str, optional): A prefix to go before every noun
                in this language (if any).
            noun_postfix (str, optuonal): A postfix to go after every noun
                in this language (if any, usually best to avoid combining
                with `noun_prefix` or language becomes very wordy).
            vowels (str, optional): Every vowel allowed in this language.
            manual_translations (dict, optional): This allows for custom-setting
                certain words in the language to mean the same thing. It is
                on the form `{real_word: fictional_word}`, for example
                `{"the", "y'e"}` .
            auto_translations (str or list, optional): These are lists
                words that should be auto-translated with a random, but
                fixed, translation. If a path to a file, this file should
                contain a list of words to produce translations for, one
                word per line.  If a list, the list's elements should be
                the words to translate.  The `manual_translations` will
                always override overlapping translations created
                automatically.
            force (bool, optional): Unless true, will not allow the addition
                of a language that is already created.

        Raises:
            LanguageExistsError: Raised if trying to adding a language
                with a key that already exists, without `force` being set.
        Notes:
            The `word_file` is for example a word-frequency list for
            the N most common words in the host language. The
            translations will be random, but will be stored
            persistently to always be the same. This allows for
            building a quick, decently-sounding fictive language that
            tend to produce the same "translation" (mostly) with the
            same input sentence.

        """
        if key in self.db.language_storage and not force:
            raise LanguageExistsError(
                "Language is already created. Re-adding it will re-build"
                " its dictionary map. Use 'force=True' keyword if you are sure."
            )

        # create grammar_component->phoneme mapping
        # {"vv": ["ea", "oh", ...], ...}
        grammar2phonemes = defaultdict(list)
        for phoneme in phonemes.split():
            if re.search(r"\W", phoneme, re.U):
                raise LanguageError("The phoneme '%s' contains an invalid character." % phoneme)
            gram = "".join(["v" if char in vowels else "c" for char in phoneme])
            grammar2phonemes[gram].append(phoneme)

        # allowed grammar are grouped by length
        gramdict = defaultdict(list)
        for gram in grammar.split():
            if re.search(r"\W|(!=[cv])", gram):
                raise LanguageError(
                    "The grammar '%s' is invalid (only 'c' and 'v' are allowed)" % gram
                )
            gramdict[len(gram)].append(gram)
        grammar = dict(gramdict)

        # create automatic translation
        translation = {}

        if auto_translations:
            if isinstance(auto_translations, str):
                # path to a file rather than a list
                with open(auto_translations, "r") as f:
                    auto_translations = f.readlines()
            for word in auto_translations:
                word = word.strip()
                lword = len(word)
                new_word = ""
                wlen = max(0, lword + sum(randint(-1, 1) for i in range(word_length_variance)))
                if wlen not in grammar:
                    # always create a translation, use random length
                    structure = choice(grammar[choice(list(grammar))])
                else:
                    # use the corresponding length
                    structure = choice(grammar[wlen])
                for match in _RE_GRAMMAR.finditer(structure):
                    try:
                        new_word += choice(grammar2phonemes[match.group()])
                    except IndexError:
                        raise IndexError(
                            "Could not find a matching phoneme for the grammar "
                            f"'{match.group()}'. Make there is at least one phoneme matching this "
                            "combination of consonants and vowels."
                        )
                translation[word.lower()] = new_word.lower()

        if manual_translations:
            # update with manual translations
            translation.update(
                dict((key.lower(), value.lower()) for key, value in manual_translations.items())
            )

        # store data
        storage = {
            "translation": translation,
            "grammar": grammar,
            "grammar2phonemes": dict(grammar2phonemes),
            "word_length_variance": word_length_variance,
            "noun_translate": noun_translate,
            "noun_prefix": noun_prefix,
            "noun_postfix": noun_postfix,
        }
        self.db.language_storage[key] = storage

    def _translate_sub(self, match):
        """
        Replacer method called by re.sub when
        traversing the language string.

        Args:
            match (re.matchobj): Match object from regex.

        Returns:
            converted word.
        Notes:
            Assumes self.lastword and self.level is available
            on the object.

        """
        word = match.group()
        lword = len(word)

        # find out what preceeded this word
        wpos = match.start()
        preceeding = match.string[:wpos].strip()
        start_sentence = preceeding.endswith((".", "!", "?")) or not preceeding

        if len(word) <= self.level:
            # below level. Don't translate
            new_word = word
        else:
            # try to translate the word from dictionary
            new_word = self.language["translation"].get(word.lower(), "")
            if not new_word:
                # no dictionary translation. Generate one

                # make up translation on the fly. Length can
                # vary from un-translated word.
                wlen = max(
                    0,
                    lword
                    + sum(randint(-1, 1) for i in range(self.language["word_length_variance"])),
                )
                grammar = self.language["grammar"]
                if wlen not in grammar:
                    if randint(0, 1) == 0:
                        # this word has no direct translation!
                        wlen = 0
                        new_word = ""
                    else:
                        # use random word length
                        wlen = choice(list(grammar.keys()))

                if wlen:
                    structure = choice(grammar[wlen])
                    grammar2phonemes = self.language["grammar2phonemes"]
                    for match in _RE_GRAMMAR.finditer(structure):
                        # there are only four combinations: vv,cc,c,v
                        try:
                            new_word += choice(grammar2phonemes[match.group()])
                        except KeyError:
                            logger.log_trace(
                                "You need to supply at least one example of each of "
                                "the four base phonemes (c, v, cc, vv)"
                            )
                            # abort translation here
                            new_word = ""
                            break

                if word.istitle():
                    if not start_sentence:
                        # this is a noun. We miss nouns at the start of
                        # sentences this way, but it's as good as we can get
                        # with this simple analysis. Maybe the fantasy language
                        # just don't consider nouns at the beginning of
                        # sentences, who knows?
                        if not self.language.get("noun_translate", False):
                            # don't translate what we identify as proper nouns (names)
                            new_word = word

                        # add noun prefix and/or postfix
                        new_word = "{prefix}{word}{postfix}".format(
                            prefix=self.language["noun_prefix"],
                            word=new_word.capitalize(),
                            postfix=self.language["noun_postfix"],
                        )

            if len(word) > 1 and word.isupper():
                # keep LOUD words loud also when translated
                new_word = new_word.upper()

            if start_sentence:
                new_word = new_word.capitalize()

        return new_word

    def translate(self, text, level=0.0, language="default"):
        """
        Translate the text according to the given level.

        Args:
            text (str): The text to translate
            level (real): Value between 0.0 and 1.0, where
                0.0 means no obfuscation (text returned unchanged) and
                1.0 means full conversion of every word. The closer to
                1, the shorter words will be translated.
            language (str): The language key identifier.

        Returns:
            text (str): A translated string.

        """
        if level == 0.0:
            # no translation
            return text
        language = self.db.language_storage.get(language, None)
        if not language:
            return text
        self.language = language

        # configuring the translation
        self.level = int(10 * (1.0 - max(0, min(level, 1.0))))
        translation = _RE_WORD.sub(self._translate_sub, text)
        # the substitution may create too long empty spaces, remove those
        return _RE_EXTRA_CHARS.sub("", translation)


# Language access functions

_LANGUAGE_HANDLER = None


def obfuscate_language(text, level=0.0, language="default"):
    """
    Main access method for the language parser.

    Args:
        text (str): Text to obfuscate.
        level (real, optional): A value from 0.0-1.0 determining
            the level of obfuscation where 0 means no obfuscation
            (string returned unchanged) and 1.0 means the entire
            string is obfuscated.
        language (str, optional): The identifier of a language
            the system understands.

    Returns:
        translated (str): The translated text.

    """
    # initialize the language handler and cache it
    global _LANGUAGE_HANDLER
    if not _LANGUAGE_HANDLER:
        try:
            _LANGUAGE_HANDLER = LanguageHandler.objects.get(db_key="language_handler")
        except LanguageHandler.DoesNotExist:
            if not _LANGUAGE_HANDLER:
                from evennia import create_script

                _LANGUAGE_HANDLER = create_script(LanguageHandler)
    return _LANGUAGE_HANDLER.translate(text, level=level, language=language)


def add_language(**kwargs):
    """
    Access function to creating a new language. See the docstring of
    `LanguageHandler.add` for list of keyword arguments.

    """
    global _LANGUAGE_HANDLER
    if not _LANGUAGE_HANDLER:
        try:
            _LANGUAGE_HANDLER = LanguageHandler.objects.get(db_key="language_handler")
        except LanguageHandler.DoesNotExist:
            if not _LANGUAGE_HANDLER:
                from evennia import create_script

                _LANGUAGE_HANDLER = create_script(LanguageHandler)
    _LANGUAGE_HANDLER.add(**kwargs)


def available_languages():
    """
    Returns all available language keys.

    Returns:
        languages (list): List of key strings of all available
        languages.
    """
    global _LANGUAGE_HANDLER
    if not _LANGUAGE_HANDLER:
        try:
            _LANGUAGE_HANDLER = LanguageHandler.objects.get(db_key="language_handler")
        except LanguageHandler.DoesNotExist:
            if not _LANGUAGE_HANDLER:
                from evennia import create_script

                _LANGUAGE_HANDLER = create_script(LanguageHandler)
    return list(_LANGUAGE_HANDLER.attributes.get("language_storage", {}))


# -----------------------------------------------------------------------------
#
# Whisper obscuration
#
# This obsucration table is designed by obscuring certain vowels first,
# following by consonants that tend to be more audible over long distances,
# like s. Finally it does non-auditory replacements, like exclamation marks and
# capitalized letters (assumed to be spoken louder) that may still give a user
# some idea of the sentence structure. Then the  word lengths are also
# obfuscated and finally the whisper length itself.
#
# ------------------------------------------------------------------------------


_RE_WHISPER_OBSCURE = [
    re.compile(r"^$", _RE_FLAGS),  # This is a Test! #0 full whisper
    re.compile(r"[ae]", _RE_FLAGS),  # This -s - Test! #1 add uy
    re.compile(r"[aeuy]", _RE_FLAGS),  # This -s - Test! #2 add oue
    re.compile(r"[aeiouy]", _RE_FLAGS),  # Th-s -s - T-st! #3 add all consonants
    re.compile(r"[aeiouybdhjlmnpqrv]", _RE_FLAGS),  # T--s -s - T-st! #4 add hard consonants
    re.compile(r"[a-eg-rt-z]", _RE_FLAGS),  # T--s -s - T-s-! #5 add all capitals
    re.compile(r"[A-EG-RT-Za-eg-rt-z]", _RE_FLAGS),  # ---s -s - --s-! #6 add f
    re.compile(r"[A-EG-RT-Za-rt-z]", _RE_FLAGS),  # ---s -s - --s-! #7 add s
    re.compile(r"[A-EG-RT-Za-z]", _RE_FLAGS),  # ---- -- - ----! #8 add capital F
    re.compile(r"[A-RT-Za-z]", _RE_FLAGS),  # ---- -- - ----! #9 add capital S
    re.compile(r"[\w]", _RE_FLAGS),  # ---- -- - ----! #10 non-alphanumerals
    re.compile(r"[\S]", _RE_FLAGS),  # ---- -- - ----  #11 words
    re.compile(r"[\w\W]", _RE_FLAGS),  # --------------  #12 whisper length
    re.compile(r".*", _RE_FLAGS),
]  # ...             #13 (always same length)


def obfuscate_whisper(whisper, level=0.0):
    """
    Obfuscate whisper depending on a pre-calculated level
    (that may depend on distance, listening skill etc)

    Args:
        whisper (str): The whisper string to obscure. The
            entire string will be considered in the obscuration.
        level (real, optional): This is a value 0-1, where 0
            means not obscured (whisper returned unchanged) and 1
            means fully obscured.

    """
    level = min(max(0.0, level), 1.0)
    olevel = int(13.0 * level)
    if olevel == 13:
        return "..."
    else:
        return _RE_WHISPER_OBSCURE[olevel].sub("-", whisper)



"""
Tests of create functions

"""

from django.test import TestCase

from evennia.scripts.scripts import DefaultScript
from evennia.utils import create
from evennia.utils.test_resources import BaseEvenniaTest


class TestCreateScript(BaseEvenniaTest):
    def test_create_script(self):
        class TestScriptA(DefaultScript):
            def at_script_creation(self):
                self.key = "test_script"
                self.interval = 10
                self.persistent = False

        script = create.create_script(TestScriptA, key="test_script")
        assert script is not None
        assert script.interval == 10
        assert script.key == "test_script"
        script.stop()

    def test_create_script_w_repeats_equal_1(self):
        class TestScriptB(DefaultScript):
            def at_script_creation(self):
                self.key = "test_script"
                self.interval = 10
                self.repeats = 1
                self.persistent = False

        # script should still exist even though repeats=1, start_delay=False
        script = create.create_script(TestScriptB, key="test_script")
        assert script
        # but the timer should be inactive now
        assert not script.is_active

    def test_create_script_w_repeats_equal_1_persisted(self):
        class TestScriptB1(DefaultScript):
            def at_script_creation(self):
                self.key = "test_script"
                self.interval = 10
                self.repeats = 1
                self.persistent = True

        # script is already stopped (interval=1, start_delay=False)
        script = create.create_script(TestScriptB1, key="test_script")
        assert script
        assert not script.is_active

    def test_create_script_w_repeats_equal_2(self):
        class TestScriptC(DefaultScript):
            def at_script_creation(self):
                self.key = "test_script"
                self.interval = 10
                self.repeats = 2
                self.persistent = False

        script = create.create_script(TestScriptC, key="test_script")
        assert script is not None
        assert script.interval == 10
        assert script.repeats == 2
        assert script.key == "test_script"
        script.stop()

    def test_create_script_w_repeats_equal_1_and_delayed(self):
        class TestScriptD(DefaultScript):
            def at_script_creation(self):
                self.key = "test_script"
                self.interval = 10
                self.start_delay = True
                self.repeats = 1
                self.persistent = False

        script = create.create_script(TestScriptD, key="test_script")
        assert script is not None
        assert script.interval == 10
        assert script.repeats == 1
        assert script.key == "test_script"
        script.stop()


class TestCreateHelpEntry(TestCase):
    help_entry = """
    Qui laborum voluptas quis commodi ipsum quo temporibus eum. Facilis
    assumenda facilis architecto in corrupti. Est placeat eum amet qui beatae
    reiciendis. Accusamus vel aspernatur ab ex. Quam expedita sed expedita
    consequuntur est dolorum non exercitationem.

    Ipsa vel ut dolorem voluptatem adipisci velit. Sit odit temporibus mollitia
    illum ipsam placeat. Rem et ipsum dolor. Hic eum tempore excepturi qui veniam
    magni.

    Excepturi quam repellendus inventore excepturi fugiat quo quasi molestias.
    Nostrum ut assumenda enim a. Repellat quis omnis est officia accusantium. Fugit
    facere qui aperiam. Perspiciatis commodi dolores ipsam nemo consequatur
    quisquam qui non. Adipisci et molestias voluptatum est sed fugiat facere.

    """

    def test_create_help_entry__simple(self):
        entry = create.create_help_entry("testentry", self.help_entry, category="Testing")
        self.assertEqual(entry.key, "testentry")
        self.assertEqual(entry.entrytext, self.help_entry)
        self.assertEqual(entry.help_category, "Testing")

        # creating same-named entry should not work (must edit existing)
        self.assertFalse(create.create_help_entry("testentry", "testtext"))

    def test_create_help_entry__complex(self):
        locks = "foo:false();bar:true()"
        aliases = ["foo", "bar", "tst"]
        tags = [("tag1", "help"), ("tag2", "help"), ("tag3", "help")]

        entry = create.create_help_entry(
            "testentry",
            self.help_entry,
            category="Testing",
            locks=locks,
            aliases=aliases,
            tags=tags,
        )
        self.assertTrue(all(lock in entry.locks.all() for lock in locks.split(";")))
        self.assertEqual(list(entry.aliases.all()).sort(), aliases.sort())
        self.assertEqual(entry.tags.all(return_key_and_category=True), tags)


class TestCreateMessage(BaseEvenniaTest):
    msgtext = """
    Qui laborum voluptas quis commodi ipsum quo temporibus eum. Facilis
    assumenda facilis architecto in corrupti. Est placeat eum amet qui beatae
    reiciendis. Accusamus vel aspernatur ab ex. Quam expedita sed expedita
    consequuntur est dolorum non exercitationem.
    """

    def test_create_msg__simple(self):
        # from evennia import set_trace;set_trace()
        msg = create.create_message(self.char1, self.msgtext, header="TestHeader")
        msg.senders = "ExternalSender"
        msg.receivers = self.char2
        msg.receivers = "ExternalReceiver"
        self.assertEqual(msg.message, self.msgtext)
        self.assertEqual(msg.header, "TestHeader")
        self.assertEqual(msg.senders, [self.char1, "ExternalSender"])
        self.assertEqual(msg.receivers, [self.char2, "ExternalReceiver"])

    def test_create_msg__custom(self):
        locks = "foo:false();bar:true()"
        tags = ["tag1", "tag2", "tag3"]
        msg = create.create_message(
            self.char1,
            self.msgtext,
            header="TestHeader",
            receivers=[self.char1, self.char2, "ExternalReceiver"],
            locks=locks,
            tags=tags,
        )
        self.assertEqual(set(msg.receivers), set([self.char1, self.char2, "ExternalReceiver"]))
        self.assertTrue(all(lock in msg.locks.all() for lock in locks.split(";")))
        self.assertEqual(msg.tags.all(), tags)


class TestCreateChannel(TestCase):
    def test_create_channel__simple(self):
        chan = create.create_channel("TestChannel1", desc="Testing channel")
        self.assertEqual(chan.key, "TestChannel1")
        self.assertEqual(chan.db.desc, "Testing channel")

    def test_create_channel__complex(self):
        locks = "foo:false();bar:true()"
        tags = ["tag1", "tag2", "tag3"]
        aliases = ["foo", "bar", "tst"]

        chan = create.create_channel(
            "TestChannel2", desc="Testing channel", aliases=aliases, locks=locks, tags=tags
        )
        self.assertTrue(all(lock in chan.locks.all() for lock in locks.split(";")))
        self.assertEqual(chan.tags.all(), tags)
        self.assertEqual(list(chan.aliases.all()).sort(), aliases.sort())



"""
This file is meant for when you want to share your game dir with
others but don't want to share all details of your specific game
or local server setup. The settings in this file will override those
in settings.py and is in .gitignore by default.

A good guideline when sharing your game dir is that you want your
game to run correctly also without this file and only use this
to override your public, shared settings.

"""

# The secret key is randomly seeded upon creation. It is used to sign
# Django's cookies and should not be publicly known. It should also
# generally not be changed once people have registered with the game
# since it will invalidate their existing sessions.
SECRET_KEY = {secret_key}



import random

from .buff import BaseBuff, Mod


class Exploit(BaseBuff):
    key = "exploit"
    name = "Exploit"
    flavor = "You are learning your opponent's weaknesses."

    duration = -1
    maxstacks = 20

    triggers = ["hit"]

    stack_msg = {
        1: "You begin to notice flaws in your opponent's defense.",
        10: "You've begun to match the battle's rhythm.",
        20: "You've found a gap in the guard!",
    }

    def conditional(self, *args, **kwargs):
        if self.handler.get_by_type(Exploited):
            return False
        return True

    def at_trigger(self, trigger: str, *args, **kwargs):
        chance = self.stacks / 20
        roll = random.random()

        if chance > roll:
            self.handler.add(Exploited)
            self.owner.msg("An opportunity presents itself!")
        elif chance < roll:
            self.handler.add(Exploit)

        if self.stacks in self.stack_msg:
            self.owner.msg(self.stack_msg[self.stacks])


class Exploited(BaseBuff):
    key = "exploited"
    name = "Exploited"
    flavor = "You have sensed your target's vulnerability, and are poised to strike."

    duration = 30

    mods = [Mod("damage", "add", 100)]

    def at_post_check(self, *args, **kwargs):
        self.owner.msg("You ruthlessly exploit your target's weakness!")
        self.remove(quiet=True)

    def at_remove(self, *args, **kwargs):
        self.owner.msg("You have waited too long; the opportunity passes.")


class Leeching(BaseBuff):
    key = "leeching"
    name = "Leeching"
    flavor = "Attacking this target fills you with vigor."

    duration = 30

    triggers = ["taken_damage"]

    def at_trigger(self, trigger: str, attacker=None, damage=None, *args, **kwargs):
        if not attacker or not damage:
            return
        attacker.msg("You have been healed for {heal} life!".format(heal=damage * 0.1))


class Poison(BaseBuff):
    key = "poison"
    name = "Poison"
    flavor = "A poison wracks this body with painful spasms."

    duration = 120

    maxstacks = 5
    tickrate = 5
    dmg = 5

    playtime = True

    def at_pause(self, *args, **kwargs):
        self.owner.db.prelogout_location.msg_contents(
            "{actor} stops twitching, their flesh a deathly pallor.".format(actor=self.owner)
        )

    def at_unpause(self, *args, **kwargs):
        self.owner.location.msg_contents(
            "{actor} begins to twitch again, their cheeks flushing red with blood.".format(
                actor=self.owner
            )
        )

    def at_tick(self, initial=True, *args, **kwargs):
        _dmg = self.dmg * self.stacks
        if not initial:
            self.owner.location.msg_contents(
                "Poison courses through {actor}'s body, dealing {damage} damage.".format(
                    actor=self.owner, damage=_dmg
                )
            )


class Sated(BaseBuff):
    key = "sated"
    name = "Sated"
    flavor = "You have eaten a great meal!"

    duration = 180
    maxstacks = 3

    mods = [Mod("mood", "add", 15)]


class StatBuff(BaseBuff):
    """Customize the stat this buff affects by feeding a list in the order [stat, mod, base, perstack] to the cache argument when added"""

    key = "statbuff"
    name = "statbuff"
    flavor = "This buff affects the following stats: {stats}"

    maxstacks = 0
    refresh = True
    unique = False

    cache = {"modgen": ["foo", "add", 0, 0]}

    def __init__(self, handler, buffkey, cache={}) -> None:
        super().__init__(handler, buffkey, cache)
        # Finds our "modgen" cache value, which we pass on application
        modgen = list(self.cache.get("modgen"))
        if modgen:
            self.mods = [Mod(*modgen)]
        msg = ""
        _msg = [mod.stat for mod in self.mods]
        for stat in _msg:
            msg += stat
        self.flavor = self.flavor.format(stats=msg)
